#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

DEFAULT = {
    'arp_disable': False,
    'default_virtual_server_vrid': None,
    'logging_template': None,
    'policy_template': None,
    'template_virtual_server': None,
    'default_virtual_server_vrid': None,
    'ipinip': False,
    'no_dest_nat': None,
    'ha_conn_mirror': None,
    'template_virtual_port': None,
    'template_tcp': None,
    'template_policy': None,
    'autosnat': True,
    'conn_limit': None,
    'template_http': None,
    'templates': None,
    'conn_limit': None,
    'conn_resume': None,
    'templates': None,
    'use_rcv_hop_for_resp': False
}



"""add_vrrp_set_table

Revision ID: cfce01f4010d
Revises: a3c90b57952d
Create Date: 2021-04-26 05:34:36.527145

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'cfce01f4010d'
down_revision = 'a3c90b57952d'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('vrrp_set',
    sa.Column('mgmt_subnet', sa.String(length=64), nullable=False),
    sa.Column('project_id', sa.String(length=64), nullable=False),
    sa.Column('set_id', sa.Integer(), nullable=False),
    sa.UniqueConstraint('mgmt_subnet', 'project_id', name='unique_name_project_subnet_set_id')
    )


def downgrade():
    op.drop_table('vrrp_set')



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from requests.exceptions import ConnectionError
from taskflow import task

import acos_client.errors as acos_errors

from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator_for_revert

LOG = logging.getLogger(__name__)


class NatPoolCreate(task.Task):
    """Task to create nat pool"""
    def __init__(self, **kwargs):
        super(NatPoolCreate, self).__init__(**kwargs)
        self._added_pool_list = []

    @axapi_client_decorator
    def execute(self, loadbalancer, vthunder, flavor_data=None):
        device_pool = None
        if flavor_data:
            natpool_flavor_list = flavor_data.get('nat_pool_list')
            natpool_flavor = flavor_data.get('nat_pool')
            if natpool_flavor_list:
                for nat_pool in natpool_flavor_list:
                    device_pool = self.axapi_client.nat.pool.try_get(nat_pool['pool_name'])
                    if not (device_pool and
                            (device_pool['pool']['start-address'] == nat_pool['start_address'] and
                                device_pool['pool']['end-address'] == nat_pool['end_address'])):
                        try:
                            self.axapi_client.nat.pool.create(**nat_pool)
                            self._added_pool_list.append(nat_pool['pool_name'])
                        except(acos_errors.Exists) as e:
                            LOG.exception("Nat-pool with name %s already exists on partition %s of "
                                          "thunder device %s",
                                          nat_pool['pool_name'], vthunder.partition_name,
                                          vthunder.ip_address)
                            raise e
                        except Exception as e:
                            LOG.exception("Failed to create nat-pool with name %s on partition %s"
                                          " of thunder device %s",
                                          nat_pool['pool_name'], vthunder.partition_name,
                                          vthunder.ip_address)
                            raise e
            if natpool_flavor:
                device_pool = self.axapi_client.nat.pool.try_get(natpool_flavor['pool_name'])
                if (device_pool and
                    (device_pool['pool']['start-address'] == natpool_flavor['start_address']
                        and device_pool['pool']['end-address'] == natpool_flavor['end_address']
                     )):
                    return
                try:
                    self.axapi_client.nat.pool.create(**natpool_flavor)
                    self._added_pool_list.append(natpool_flavor['pool_name'])
                except(acos_errors.Exists) as e:
                    LOG.exception("Nat-pool with name %s already exists on partition %s of "
                                  "thunder device %s",
                                  natpool_flavor['pool_name'], vthunder.partition_name,
                                  vthunder.ip_address)
                    raise e
                except Exception as e:
                    LOG.exception("Failed to create nat-pool with name %s on partition %s of "
                                  "thunder device %s",
                                  natpool_flavor['pool_name'], vthunder.partition_name,
                                  vthunder.ip_address)
                    raise e

    @axapi_client_decorator_for_revert
    def revert(self, loadbalancer, vthunder, flavor_data=None, *args, **kwargs):
        if flavor_data:
            natpool_flavor_list = flavor_data.get('nat_pool_list')
            natpool_flavor = flavor_data.get('nat_pool')
            try:
                if len(self._added_pool_list) > 0:
                    if natpool_flavor_list:
                        for nat_pool in natpool_flavor_list:
                            pool_name = nat_pool.get('pool_name')
                            if pool_name in self._added_pool_list:
                                LOG.warning("Reverting creation of nat-pool: %s", pool_name)
                                self.axapi_client.nat.pool.delete(pool_name)
                    if natpool_flavor:
                        pool_name = natpool_flavor.get('pool_name')
                        if pool_name in self._added_pool_list:
                            LOG.warning("Reverting creation of nat-pool: %s", pool_name)
                            self.axapi_client.nat.pool.delete(pool_name)
            except ConnectionError:
                LOG.exception(
                    "Failed to connect A10 Thunder device: %s", vthunder.ip_address)
            except (acos_errors.ACOSException, Exception) as e:
                LOG.exception("Failed to revert creation of nat-pool %s due to: %s",
                              pool_name, str(e))


class NatPoolDelete(task.Task):
    """Task to delete nat pool"""

    @axapi_client_decorator
    def execute(self, loadbalancer, vthunder, lb_count_flavor, flavor_data=None):
        if vthunder:
            if lb_count_flavor <= 1:
                if flavor_data:
                    natpool_flavor_list = flavor_data.get('nat_pool_list')
                    natpool_flavor = flavor_data.get('nat_pool')
                    if natpool_flavor_list:
                        for nat_pool in natpool_flavor_list:
                            pool_name = nat_pool['pool_name']
                            try:
                                self.axapi_client.nat.pool.delete(pool_name)
                            except(acos_errors.ACOSException) as e:
                                LOG.exception("Failed to delete Nat-pool with name %s due to %s",
                                              pool_name, str(e))
                    if natpool_flavor:
                        pool_name = natpool_flavor['pool_name']
                        try:
                            self.axapi_client.nat.pool.delete(pool_name)
                        except(acos_errors.ACOSException) as e:
                            LOG.exception("Failed to delete Nat-pool with name %s due to %s",
                                          pool_name, str(e))
            else:
                LOG.warning("Cannot delete Nat-pool(s) in flavor %s as "
                            "they are in use by another loadbalancer(s)", loadbalancer.flavor_id)



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from contextlib import contextmanager

import sqlalchemy
import sqlalchemy.ext.declarative
import sqlalchemy.orm
from a10_octavia import a10_config

A10_CFG = None
Base = sqlalchemy.ext.declarative.declarative_base()


def get_base():
    return Base


def get_engine(url=None):
    global A10_CFG

    if url is None:
        if A10_CFG is None:
            A10_CFG = a10_config.A10Config()

        url = A10_CFG.get('database_connection')

    return sqlalchemy.create_engine(url)


def get_session(url=None, **kwargs):
    DBSession = sqlalchemy.orm.sessionmaker(bind=get_engine(url=url))
    return DBSession(**kwargs)


def close_session(session):
    try:
        session.commit()
    finally:
        session.close()


@contextmanager
def magic_session(db_session=None, url=None):
    """Either does nothing with the session you already have or
    makes one that commits and closes no matter what happens
    """

    if db_session is not None:
        yield db_session
    else:
        session = get_session(url, expire_on_commit=False)
        try:
            try:
                yield session
            finally:
                session.commit()
        finally:
            session.close()



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
import imp
try:
    from unittest import mock
except ImportError:
    import mock
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from octavia.common import data_models as o_data_models
from octavia.network import data_models as o_net_data_models

from a10_octavia.common import config_options
from a10_octavia.common import data_models
from a10_octavia.controller.worker.tasks import a10_network_tasks
from a10_octavia.tests.common import a10constants
from a10_octavia.tests.unit import base

MEMBER = o_data_models.Member(subnet_id=a10constants.MOCK_SUBNET_ID)
POOL = o_data_models.Pool()
VTHUNDER = data_models.VThunder()
SUBNET = o_net_data_models.Subnet()
PORT = o_net_data_models.Port()
VRID = data_models.VRID()
VRID_VALUE = 0
SUBNET_1 = o_net_data_models.Subnet(id=a10constants.MOCK_SUBNET_ID)
VRID_1 = data_models.VRID(id=1, subnet_id=a10constants.MOCK_SUBNET_ID)
NAT_POOL = data_models.NATPool(port_id=a10constants.MOCK_PORT_ID)
NAT_FLAVOR = {"pool_name": "p1", "start_address": "1.1.1.1", "end_address": "1.1.1.2"}
HW_THUNDER = data_models.HardwareThunder(
    project_id=a10constants.MOCK_PROJECT_ID,
    device_name="rack_thunder_1",
    undercloud=True,
    username="abc",
    password="abc",
    ip_address="10.10.10.10",
    partition_name="shared")
HW_THUNDER2 = data_models.HardwareThunder(
    project_id=a10constants.MOCK_PROJECT_ID,
    device_name="rack_thunder_2",
    undercloud=True,
    username="abc",
    password="abc",
    ip_address="10.10.10.11",
    partition_name="shared",
    vrid_floating_ip="192.168.8.126")
EXISTING_FIP_SHARED_PARTITION = {
    u'vrid': {u'blade-parameters': {
        u'priority': 150, u'uuid': u'41e54b26-bc4f-11eb-bd71-525400895118',
        u'a10-url': u'/axapi/v3/vrrp-a/vrid/0/blade-parameters'},
        u'uuid': u'41e5439c-bc4f-11eb-bd71-525400895118', u'floating-ip': {
            u'ip-address-cfg': [{
                u'ip-address': u'192.168.8.140'}, {u'ip-address': u'192.168.9.140'}]},
            u'vrid-val': 0, u'preempt-mode': {u'threshold': 0, u'disable': 0},
            u'a10-url': u'/axapi/v3/vrrp-a/vrid/0'}
}
EXISTING_FIP_L3V_PARTITION = {
    u'vrid': {u'blade-parameters': {
        u'priority': 150, u'uuid': u'41e54b26-bc4f-11eb-bd71-525400895118',
        u'a10-url': u'/axapi/v3/vrrp-a/vrid/0/blade-parameters'},
        u'uuid': u'41e5439c-bc4f-11eb-bd71-525400895118', u'floating-ip': {
            u'ip-address-part-cfg': [{
                u'ip-address-partition': u'192.168.8.140'},
                {u'ip-address-partition': u'192.168.9.140'}]},
            u'vrid-val': 0, u'preempt-mode': {u'threshold': 0, u'disable': 0},
            u'a10-url': u'/axapi/v3/vrrp-a/vrid/0'}
}


class MockIP(object):
    def __init__(self, ip_address):
        self.ip_address = ip_address


class TestNetworkTasks(base.BaseTaskTestCase):

    def setUp(self):
        super(TestNetworkTasks, self).setUp()
        imp.reload(a10_network_tasks)
        patcher = mock.patch(
            'a10_octavia.controller.worker.tasks.a10_network_tasks.BaseNetworkTask.network_driver')
        self.network_driver_mock = patcher.start()
        self.client_mock = mock.Mock()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        self.conf.register_opts(config_options.A10_GLOBAL_OPTS,
                                group=a10constants.A10_GLOBAL_OPTS)

    def tearDown(self):
        super(TestNetworkTasks, self).tearDown()
        self.conf.reset()

    @mock.patch(
        'a10_octavia.common.utils.get_vrid_floating_ip_for_project',
        return_value=None)
    def test_HandleVRIDFloatingIP_noop_vrrpa_config_not_specified(
            self, mock_utils):
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        vthunder_config = copy.deepcopy(HW_THUNDER)
        result = mock_network_task.execute(VTHUNDER, MEMBER, [], subnet, vthunder_config)
        self.assertEqual(result, [])

    @mock.patch('a10_octavia.common.utils.get_vrid_floating_ip_for_project',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    @mock.patch('a10_octavia.common.utils.get_patched_ip_address',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    def test_HandleVRIDFloatingIP_create_floating_ip_in_shared_partition_with_static_ip(
            self, mock_patched_ip, mock_floating_ip):
        member = copy.deepcopy(MEMBER)
        member.subnet_id = SUBNET_1.id
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port = copy.deepcopy(PORT)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            mock.ANY, None, mock.ANY,
            fixed_ip=a10constants.MOCK_VRID_FLOATING_IP_1)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[a10constants.MOCK_VRID_FLOATING_IP_1],
            is_partition=False)

    @mock.patch('a10_octavia.common.utils.get_patched_ip_address',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    def test_HandleVRIDFloatingIP_create_floating_ip_with_device_name_flavor(
            self, mock_patched_ip):
        member = copy.deepcopy(MEMBER)
        member.subnet_id = SUBNET_1.id
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        vthunder_config = copy.deepcopy(HW_THUNDER2)
        port = copy.deepcopy(PORT)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [], subnet,
                                  vthunder_config, use_device_flavor=True)
        mock_network_task.axapi_client.get_vrid_floating_ip_for_project.assert_not_called()
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[a10constants.MOCK_VRID_FLOATING_IP_1],
            is_partition=False)

    @mock.patch('a10_octavia.common.utils.get_vrid_floating_ip_for_project',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    @mock.patch('a10_octavia.common.utils.get_patched_ip_address',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_HandleVRIDFloatingIP_create_floating_ip_in_specified_partition_with_static_ip(
            self, mock_utils, mock_patched_ip, get_floating_ip):
        member = copy.deepcopy(MEMBER)
        member.subnet_id = SUBNET_1.id
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port = copy.deepcopy(PORT)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.partition_name = 'partition_1'
        vthunder.ip_address = '10.0.0.1'
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_L3V_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            mock.ANY, None, mock.ANY,
            fixed_ip=a10constants.MOCK_VRID_FLOATING_IP_1)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[
                a10constants.MOCK_VRID_FLOATING_IP_1], is_partition=True)

    @mock.patch(
        'a10_octavia.common.utils.check_ip_in_subnet_range',
        return_value=False)
    @mock.patch(
        'a10_octavia.common.utils.get_vrid_floating_ip_for_project',
        return_value='dhcp')
    def test_HandleVRIDFloatingIP_create_floating_ip_in_shared_partition_with_dhcp(
            self, get_floating_ip, check_subnet):
        member = copy.deepcopy(MEMBER)
        member.subnet_id = SUBNET_1.id
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port = copy.deepcopy(PORT)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [VRID_1], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            mock.ANY, None, None, fixed_ip=None)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[a10constants.MOCK_VRID_FLOATING_IP_1],
            is_partition=False)

    @mock.patch(
        'a10_octavia.common.utils.check_ip_in_subnet_range',
        return_value=False)
    @mock.patch(
        'a10_octavia.common.utils.get_vrid_floating_ip_for_project',
        return_value='dhcp')
    def test_HandleVRIDFloatingIP_create_floating_ip_in_specified_partition_with_dhcp(
            self, get_floating_ip, check_subnet):
        member = copy.deepcopy(MEMBER)
        member.subnet_id = SUBNET_1.id
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port = copy.deepcopy(PORT)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.partition_name = 'partition_1'
        vthunder.ip_address = '10.0.0.1'
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_L3V_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [VRID_1], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            mock.ANY, None, None, fixed_ip=None)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[
                a10constants.MOCK_VRID_FLOATING_IP_1], is_partition=True)

    @mock.patch(
        'a10_octavia.common.utils.get_vrid_floating_ip_for_project',
        return_value=None)
    def test_HandleVRIDFloatingIP_delete_fip_entries_device_fip_given_but_no_fip_in_conf(
            self, mock_utils):
        vrid = copy.deepcopy(VRID_1)
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        vrid.vrid = VRID_VALUE
        vthunder_config = copy.deepcopy(HW_THUNDER)
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        result = mock_network_task.execute(VTHUNDER, MEMBER, [vrid], SUBNET_1, vthunder_config)
        self.network_driver_mock.delete_port.assert_called_with(
            vrid.vrid_port_id)
        self.client_mock.vrrpa.update.assert_called_with(
            vrid.vrid, floating_ips=[], is_partition=False)
        self.assertEqual(result, [])

    @mock.patch('a10_octavia.common.utils.get_vrid_floating_ip_for_project',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    @mock.patch('a10_octavia.common.utils.get_patched_ip_address',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    def test_HandleVRIDFloatingIP_noop_device_fip_and_conf_fip_both_given_same_ip(
            self, mock_patched_ip, get_floating_ip):
        vrid = copy.deepcopy(VRID_1)
        vrid.vrid_floating_ip = a10constants.MOCK_VRID_FLOATING_IP_1
        vrid.vrid = VRID_VALUE
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        member = copy.deepcopy(MEMBER)
        member.subnet_id = a10constants.MOCK_SUBNET_ID
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        port = copy.deepcopy(PORT)
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        mock_network_task.execute(vthunder, member, [vrid], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_not_called()
        self.client_mock.vrrpa.update.assert_not_called()

    @mock.patch('a10_octavia.common.utils.get_vrid_floating_ip_for_project',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_2)
    @mock.patch('a10_octavia.common.utils.get_patched_ip_address',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_2)
    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_HandleVRIDFloatingIP_replace_floating_ip_in_shared_partition_with_static_ip(
            self, mock_utils, mock_patched_ip, get_floating_ip):
        vrid = copy.deepcopy(VRID_1)
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        vrid.vrid_floating_ip = a10constants.MOCK_VRID_FLOATING_IP_1
        vrid.vrid = VRID_VALUE
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        member = copy.deepcopy(MEMBER)
        member.subnet_id = a10constants.MOCK_SUBNET_ID
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        port = copy.deepcopy(PORT)
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_2))
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [vrid], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            vrid, subnet.network_id, mock.ANY,
            fixed_ip=a10constants.MOCK_VRID_FLOATING_IP_2)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[a10constants.MOCK_VRID_FLOATING_IP_2],
            is_partition=False)
        self.network_driver_mock.delete_port.assert_called_with(
            a10constants.MOCK_VRRP_PORT_ID)

    @mock.patch('a10_octavia.common.utils.get_vrid_floating_ip_for_project',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_2)
    @mock.patch('a10_octavia.common.utils.get_patched_ip_address',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_2)
    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_HandleVRIDFloatingIP_replace_floating_ip_in_specified_partition_with_static_ip(
            self, mock_utils, mock_patched_ip, get_floating_ip):
        vrid = copy.deepcopy(VRID_1)
        vrid.vrid_floating_ip = a10constants.MOCK_VRID_FLOATING_IP_1
        vrid.vrid = VRID_VALUE
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        member = copy.deepcopy(MEMBER)
        member.subnet_id = a10constants.MOCK_SUBNET_ID
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        port = copy.deepcopy(PORT)
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_2))
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        vthunder.partition_name = 'partition_1'
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_L3V_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [vrid], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            vrid, None, mock.ANY,
            fixed_ip=a10constants.MOCK_VRID_FLOATING_IP_2)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[
                a10constants.MOCK_VRID_FLOATING_IP_2], is_partition=True)
        self.network_driver_mock.delete_port.assert_called_with(
            a10constants.MOCK_VRRP_PORT_ID)

    @mock.patch(
        'a10_octavia.common.utils.get_vrid_floating_ip_for_project',
        return_value='dhcp')
    def test_HandleVRIDFloatingIP_noop_given_same_subnet_with_conf_fip_set_to_dhcp(
            self, get_floating_ip):
        vrid = copy.deepcopy(VRID_1)
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        vrid.vrid_floating_ip = a10constants.MOCK_VRID_FLOATING_IP_1
        vrid.vrid = VRID_VALUE
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        member = copy.deepcopy(MEMBER)
        member.subnet_id = a10constants.MOCK_SUBNET_ID
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        vthunder_config = copy.deepcopy(HW_THUNDER)
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        fip_port = mock_network_task.execute(vthunder, member, [vrid], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_not_called()
        self.network_driver_mock.delete_port.assert_not_called()
        self.assertEqual(fip_port, [vrid])

    @mock.patch(
        'a10_octavia.common.utils.check_ip_in_subnet_range',
        return_value=False)
    @mock.patch(
        'a10_octavia.common.utils.get_vrid_floating_ip_for_project',
        return_value='dhcp')
    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_HandleVRIDFloatingIP_replace_floating_ip_diff_subnet_in_shared_part_conf_fip_set_dhcp(
            self, mock_utils, get_floating_ip, check_subnet):
        vrid = copy.deepcopy(VRID_1)
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        vrid.vrid_floating_ip = a10constants.MOCK_VRID_FLOATING_IP_1
        vrid.vrid = VRID_VALUE
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        member = copy.deepcopy(MEMBER)
        member.subnet_id = a10constants.MOCK_SUBNET_ID
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        port = copy.deepcopy(PORT)
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))

        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [vrid], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            vrid, subnet.network_id, mock.ANY, fixed_ip=None)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[a10constants.MOCK_VRID_FLOATING_IP_1],
            is_partition=False)
        self.network_driver_mock.delete_port.assert_called_with(
            a10constants.MOCK_VRRP_PORT_ID)

    @mock.patch('a10_octavia.common.utils.check_ip_in_subnet_range',
                return_value=False)
    @mock.patch('a10_octavia.common.utils.get_vrid_floating_ip_for_project',
                return_value='dhcp')
    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_HandleVRIDFloatingIP_replace_floating_ip_diff_subnet_in_set_part_conf_fip_set_dhcp(
            self, mock_utils, get_floating_ip, check_subnet):
        vrid = copy.deepcopy(VRID_1)
        vrid.vrid_floating_ip = a10constants.MOCK_VRID_FLOATING_IP_1
        vrid.vrid = VRID_VALUE
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        member = copy.deepcopy(MEMBER)
        member.subnet_id = a10constants.MOCK_SUBNET_ID
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        port = copy.deepcopy(PORT)
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.partition_name = 'partition_1'
        vthunder.ip_address = '10.0.0.1'
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_L3V_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [vrid], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            vrid, subnet.network_id, mock.ANY, fixed_ip=None)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[
                a10constants.MOCK_VRID_FLOATING_IP_1], is_partition=True)
        self.network_driver_mock.delete_port.assert_called_with(
            a10constants.MOCK_VRRP_PORT_ID)

    @mock.patch('a10_octavia.common.utils.get_vrid_floating_ip_for_project',
                return_value=a10constants.MOCK_VRID_PARTIAL_FLOATING_IP)
    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_HandleVRIDFloatingIP_creating_floating_ip_conf_fip_is_partial(
            self, get_floating_ip, mock_utils):
        member = copy.deepcopy(MEMBER)
        member.subnet_id = a10constants.MOCK_SUBNET_ID
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        port = copy.deepcopy(PORT)
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FULL_FLOATING_IP))

        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, member, [VRID_1], subnet, vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            mock.ANY, subnet.network_id, mock.ANY,
            fixed_ip=a10constants.MOCK_VRID_FULL_FLOATING_IP)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[a10constants.MOCK_VRID_FULL_FLOATING_IP],
            is_partition=False)

    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_DeleteMemberVRIDPort_delete_vrid_ip_member_count_equals_one(self, mock_utils):
        mock_network_task = a10_network_tasks.DeleteVRIDPort()
        vrid = copy.deepcopy(VRID_1)
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        vrid.vrid = VRID_VALUE
        mock_network_task.axapi_client = self.client_mock
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        result = mock_network_task.execute(VTHUNDER, [vrid], SUBNET_1, False,
                                           0, 1, 0, 0, MEMBER)
        self.network_driver_mock.deallocate_vrid_fip.assert_called_with(
            vrid, mock.ANY, mock.ANY)
        self.client_mock.vrrpa.update.assert_called_with(
            vrid.vrid, floating_ips=[])
        self.assertEqual(result, (vrid, True))

    def test_DeleteMemberVRIDPort_member_count_lb_count(self):
        mock_network_task = a10_network_tasks.DeleteVRIDPort()
        mock_network_task.axapi_client = self.client_mock
        result = mock_network_task.execute(VTHUNDER, [VRID_1], SUBNET_1, False,
                                           1, 1, 0, 0, MEMBER)
        self.network_driver_mock.deallocate_vrid_fip.assert_not_called()
        self.client_mock.vrrpa.delete.assert_not_called()
        self.assertEqual(result, (None, False))

    def delete_multi_vrid_port_against_subnet_not_used(self):
        mock_network_task = a10_network_tasks.DeleteVRIDPort()
        vrid = copy.deepcopy(VRID_1)
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        vrid.vrid = VRID_VALUE
        mock_network_task.axapi_client = self.client_mock
        result = mock_network_task.execute(VTHUNDER, [vrid], [SUBNET_1])
        self.network_driver_mock.delete_port.assert_called_with(
            a10constants.MOCK_VRRP_PORT_ID)
        self.client_mock.vrrpa.update.assert_called_with(
            vrid.vrid, floating_ips=[], is_partiton=False)
        self.assertEqual(result, [])

    def delete_multi_vrid_port_against_all_subnet_used(self):
        mock_network_task = a10_network_tasks.DeleteVRIDPort()
        vrid = copy.deepcopy(VRID_1)
        vrid.vrid_port_id = a10constants.MOCK_VRRP_PORT_ID
        vrid.vrid = VRID_VALUE
        mock_network_task.axapi_client = self.client_mock
        result = mock_network_task.execute(VTHUNDER, [vrid], [])
        self.network_driver_mock.delete_port.assert_not_called()
        self.client_mock.vrrpa.update.assert_not_called()
        self.assertEqual(result, None)

    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_reserve_subnet_addr_for_member(self, mock_utils):
        mock_network_task = a10_network_tasks.ReserveSubnetAddressForMember()
        mock_network_task.network_driver = self.client_mock
        mock_network_task.execute(MEMBER, NAT_FLAVOR)
        self.client_mock.reserve_subnet_addresses.assert_called_with(
            MEMBER.subnet_id, ["1.1.1.1", "1.1.1.2"], mock.ANY)

    def test_release_subnet_addr_referenced(self):
        mock_network_task = a10_network_tasks.ReleaseSubnetAddressForMember()
        NAT_POOL.member_ref_count = 2
        ret_val = mock_network_task.execute(MEMBER, NAT_FLAVOR, NAT_POOL)
        self.assertEqual(ret_val, None)

    @mock.patch('a10_octavia.controller.worker.tasks.a10_network_tasks.a10_task_utils')
    def test_release_subnet_addr(self, mock_utils):
        mock_network_task = a10_network_tasks.ReleaseSubnetAddressForMember()
        mock_network_task.network_driver = self.client_mock
        NAT_POOL.member_ref_count = 1
        mock_network_task.execute(MEMBER, NAT_FLAVOR, NAT_POOL)
        self.client_mock.delete_port.assert_called_with(NAT_POOL.port_id)

    @mock.patch('a10_octavia.common.utils.get_vrid_floating_ip_for_project',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    @mock.patch('a10_octavia.common.utils.get_patched_ip_address',
                return_value=a10constants.MOCK_VRID_FLOATING_IP_1)
    def test_HandleVRIDFloatingIP_create_floating_ip_for_subnet_list(
            self, mock_patched_ip, mock_floating_ip):
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.ip_address = '10.0.0.1'
        subnet = copy.deepcopy(SUBNET_1)
        subnet.cidr = a10constants.MOCK_SUBNET_CIDR
        vthunder_config = copy.deepcopy(HW_THUNDER)
        port = copy.deepcopy(PORT)
        port.fixed_ips.append(MockIP(a10constants.MOCK_VRID_FLOATING_IP_1))
        self.client_mock.vrrpa.get.return_value = EXISTING_FIP_SHARED_PARTITION
        mock_network_task = a10_network_tasks.HandleVRIDFloatingIP()
        mock_network_task.axapi_client = self.client_mock
        self.network_driver_mock.get_subnet.return_value = subnet
        self.network_driver_mock.allocate_vrid_fip.return_value = port
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         vrid=VRID_VALUE)
        mock_network_task.execute(vthunder, POOL, [], [subnet], vthunder_config)
        self.network_driver_mock.allocate_vrid_fip.assert_called_with(
            mock.ANY, None, mock.ANY,
            fixed_ip=a10constants.MOCK_VRID_FLOATING_IP_1)
        self.client_mock.vrrpa.update.assert_called_with(
            VRID_VALUE, floating_ips=[a10constants.MOCK_VRID_FLOATING_IP_1],
            is_partition=False)

    def test_get_all_resource_subnet(self):
        mock_network_task = a10_network_tasks.GetAllResourceSubnet()
        self.network_driver_mock.get_subnet.return_value = SUBNET_1
        subnet = mock_network_task.execute([MEMBER])
        self.assertEqual(subnet, [SUBNET_1])



"""updating_vthunders_table_with_last_write_mem

Revision ID: aa1984742932
Revises: a41260fb4ccd
Create Date: 2021-01-19 14:57:39.782860

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'aa1984742932'
down_revision = 'a41260fb4ccd'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('vthunders', sa.Column('last_write_mem', sa.DateTime(), default=None, nullable=True))


def downgrade():
    op.drop_column('vthunders', 'last_write_mem')
    pass



"""Added vrid table

Revision ID: 4028e5f7a198
Revises: b02c78488551
Create Date: 2020-05-27 14:46:46.139384

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4028e5f7a198'
down_revision = 'b02c78488551'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'vrid',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('project_id', sa.String(36), nullable=False),
        sa.Column('vrid', sa.Integer, default=0, nullable=False),
        sa.Column('vrid_port_id', sa.String(36), nullable=False),
        sa.Column('vrid_floating_ip', sa.String(40), nullable=False)
    )


def downgrade():
    op.drop_table('vrid')



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_config import cfg
from oslo_log import log as logging
from taskflow.types import failure

from octavia.common import constants
from octavia.common import exceptions
from octavia.controller.worker.tasks.compute_tasks import BaseComputeTask

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class ComputeCreate(BaseComputeTask):
    """Create the compute instance for a new amphora."""

    def execute(self, amphora_id, loadbalancer=None,
                build_type_priority=constants.LB_CREATE_NORMAL_PRIORITY,
                server_group_id=None, ports=None):

        ports = ports or []
        network_ids = CONF.a10_controller_worker.amp_boot_network_list[:]
        # Injecting VIP network ID
        if loadbalancer:
            if loadbalancer.vip.network_id not in network_ids:
                network_ids.append(loadbalancer.vip.network_id)
        LOG.debug("Compute create execute for amphora with id %s", amphora_id)
        key_name = CONF.a10_controller_worker.amp_ssh_key_name

        try:
            if CONF.a10_controller_worker.build_rate_limit != -1:
                self.rate_limit.add_to_build_request_queue(
                    amphora_id, build_type_priority)

            compute_id = self.compute.build(
                name="amphora-" + amphora_id,
                amphora_flavor=CONF.a10_controller_worker.amp_flavor_id,
                image_id=CONF.a10_controller_worker.amp_image_id,
                image_tag=CONF.a10_controller_worker.amp_image_tag,
                image_owner=CONF.a10_controller_worker.amp_image_owner_id,
                key_name=key_name,
                sec_groups=CONF.a10_controller_worker.amp_secgroup_list,
                network_ids=network_ids,
                port_ids=[port.id for port in ports],
                server_group_id=server_group_id)

            LOG.debug("Server created with id: %s for amphora id: %s",
                      compute_id, amphora_id)
            return compute_id

        except Exception as e:
            LOG.error("Compute create for amphora id: %s failed",
                      amphora_id)
            raise e

    def revert(self, result, amphora_id, *args, **kwargs):
        """This method will revert the creation of the amphora"""

        if isinstance(result, failure.Failure):
            return
        compute_id = result
        LOG.warning("Reverting compute create for amphora with id"
                    "%(amp)s and compute id: %(comp)s",
                    {'amp': amphora_id, 'comp': compute_id})
        try:
            self.compute.delete(compute_id)
        except Exception as e:
            LOG.exception("Failed to revert creation of compute %s due to %s: ",
                          compute_id, str(e))


class ComputeActiveWait(BaseComputeTask):
    """Wait for the compute driver to mark the amphora active."""

    def execute(self, compute_id, amphora_id):
        """Wait for the compute driver to mark the amphora active"""
        for i in range(CONF.a10_controller_worker.amp_active_retries):
            amp, fault = self.compute.get_amphora(compute_id)
            if amp.status == constants.ACTIVE:
                if CONF.a10_controller_worker.build_rate_limit != -1:
                    self.rate_limit.remove_from_build_req_queue(amphora_id)
                return amp
            elif amp.status == constants.ERROR:
                raise exceptions.ComputeBuildException(fault=fault)
            time.sleep(CONF.a10_controller_worker.amp_active_wait_sec)

        raise exceptions.ComputeWaitTimeoutException(id=compute_id)


class NovaServerGroupCreate(BaseComputeTask):
    def execute(self, loadbalancer_id):
        """Create a server group by nova client api"""

        name = 'octavia-lb-' + loadbalancer_id
        server_group = self.compute.create_server_group(
            name, CONF.a10_nova.anti_affinity_policy)
        LOG.debug("Server Group created with id: %s  for load balancer id: %s"
                  "", server_group.id, loadbalancer_id)
        return server_group.id

    def revert(self, result, *args, **kwargs):
        """This method will revert the creation of the nova server group"""

        server_group_id = result
        LOG.warning("Reverting server group create with id: %s",
                    server_group_id)
        try:
            self.compute.delete_server_group(server_group_id)
        except Exception as e:
            LOG.error("Failed to delete server group.  Resources may "
                      "still be in use for server group: %(sg)s due to "
                      "error: %(except)s",
                      {'sg': server_group_id, 'except': e})



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import json

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from octavia.tests.unit import base

from a10_octavia.common import config_options
from a10_octavia.common import config_types
from a10_octavia.common import data_models as data_models
from a10_octavia.tests.common import a10constants

CONF = cfg.CONF

RACK_DEVICE_1 = {
    "project_id": "project-1",
    "ip_address": "10.0.0.1",
    "device_name": "rack_thunder_1",
    "username": "abc",
    "password": "abc",
    "interface_vlan_map": {
        "device_1": {
            "ethernet_interfaces": [{
                "interface_num": 5,
                "vlan_map": [
                    {"vlan_id": 11, "use_dhcp": "True"},
                    {"vlan_id": 12, "ve_ip": ".10"}
                ]
            }]
        }
    }
}

RACK_DEVICE_2 = {
    "project_id": "project-2",
    "ip_address": "11.0.0.1",
    "device_name": "rack_thunder_2",
    "username": "abc",
    "password": "abc",
    'partition_name': 'def-sample'
}

VTHUNDER_1 = data_models.HardwareThunder(project_id="project-1", device_name="rack_thunder_1",
                                         undercloud=True, username="abc", password="abc",
                                         ip_address="10.0.0.1", partition_name="shared",
                                         device_network_map='"device_1":{"ethernet_interfaces":'
                                         '[{"interface_num": 5, "vlan_map": ['
                                         '{"vlan_id": 11, "use_dhcp": "True"},'
                                         '{"vlan_id": 12, "ve_ip": ".10"}]}]}}')
VTHUNDER_2 = data_models.HardwareThunder(project_id="project-2", device_name="rack_thunder_2",
                                         undercloud=True, username="def", password="def",
                                         ip_address="11.0.0.1", partition_name="def-sample")

RESULT_RACK_DEVICE_LIST = {'project-1': VTHUNDER_1,
                           'project-2': VTHUNDER_2}


class TestConfigTypes(base.TestCase):

    def setUp(self):
        super(TestConfigTypes, self).setUp()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))

    def tearDown(self):
        super(TestConfigTypes, self).tearDown()
        self.conf.reset()

    def test_rack_device_valid_devices(self):
        self.conf.register_opts(config_options.A10_HARDWARE_THUNDER_OPTS,
                                group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION)
        devices_str = json.dumps([RACK_DEVICE_1, RACK_DEVICE_2])
        self.conf.config(group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION,
                         devices=devices_str)
        self.assertIn('project-1', CONF.hardware_thunder.devices)
        self.assertIn('project-2', CONF.hardware_thunder.devices)
        self.assertEqual('rack_thunder_1', CONF.hardware_thunder.devices['project-1'].device_name)
        self.assertEqual('rack_thunder_2', CONF.hardware_thunder.devices['project-2'].device_name)

    def test_rack_device_valid_no_devices(self):
        self.conf.register_opts(config_options.A10_HARDWARE_THUNDER_OPTS,
                                group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION)
        devices_str = json.dumps([])
        self.conf.config(group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION,
                         devices=devices_str)
        self.assertEqual(CONF.hardware_thunder.devices, {})

    def test_rack_device_valid_invalid_array(self):
        devices_str = '[' + json.dumps(RACK_DEVICE_1)
        lob_object = config_types.ListOfObjects(bounds=True)
        self.assertRaises(ValueError, lob_object, devices_str)

    def test_rack_device_valid_invalid_dict(self):
        devices_str = '[{' + json.dumps(RACK_DEVICE_1) + ']'
        lob_object = config_types.ListOfObjects(bounds=True)
        self.assertRaises(SyntaxError, lob_object, devices_str)



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client.errors as acos_errors
from oslo_config import cfg
from oslo_log import log as logging
from requests.exceptions import ConnectionError
from taskflow import task

from a10_octavia.common import a10constants
from a10_octavia.common import exceptions
from a10_octavia.common import openstack_mappings
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator_for_revert
from a10_octavia.controller.worker.tasks import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class ListenersParent(object):

    def set(self, set_method, loadbalancer, listener, vthunder, flavor_data=None,
            update_dict=None, ssl_template=None):
        listener.load_balancer = loadbalancer
        listener.protocol = openstack_mappings.virtual_port_protocol(
            self.axapi_client, listener.protocol).lower()

        config_data = {
            'ipinip': CONF.listener.ipinip,
            'use_rcv_hop': CONF.listener.use_rcv_hop_for_resp,
            'ha_conn_mirror': CONF.listener.ha_conn_mirror
        }

        status = self.axapi_client.slb.UP
        if not listener.enabled:
            status = self.axapi_client.slb.DOWN
        config_data['status'] = status

        conn_limit = CONF.listener.conn_limit
        if listener.connection_limit != -1:
            conn_limit = listener.connection_limit
        if conn_limit < 1 or conn_limit > 64000000:
            raise Exception('The specified member server connection limit '
                            '(configuration setting: conn-limit) is out of '
                            'bounds with value {0}. Please set to between '
                            '1-64000000.'.format(conn_limit))
        config_data['conn_limit'] = conn_limit

        no_dest_nat = CONF.listener.no_dest_nat
        if no_dest_nat and (
                listener.protocol
                not in a10constants.NO_DEST_NAT_SUPPORTED_PROTOCOL):
            LOG.warning("'no_dest_nat' is not allowed for HTTP," +
                        "HTTPS or TERMINATED_HTTPS listener.")
            no_dest_nat = False
        config_data['no_dest_nat'] = no_dest_nat

        autosnat = CONF.listener.autosnat
        if autosnat and no_dest_nat:
            raise exceptions.SNATConfigurationError()
        config_data['autosnat'] = autosnat

        c_pers, s_pers = utils.get_sess_pers_templates(listener.default_pool)
        device_templates = self.axapi_client.slb.template.templates.get()
        vport_templates = {}
        template_vport = CONF.listener.template_virtual_port
        if template_vport and template_vport.lower() != 'none':
            template_key = 'template-virtual-port'
            if vthunder.partition_name != "shared":
                if CONF.a10_global.use_shared_for_template_lookup:
                    template_key = utils.shared_template_modifier(template_key,
                                                                  template_vport,
                                                                  device_templates)
            vport_templates[template_key] = template_vport

        template_args = {}
        if listener.protocol.upper() in a10constants.HTTP_TYPE:
            if listener.protocol == 'https' and listener.tls_certificate_id:
                # Adding TERMINATED_HTTPS SSL cert, created in previous task
                template_args["template_client_ssl"] = listener.id

            template_http = CONF.listener.template_http
            if template_http and template_http.lower() != 'none':
                template_key = 'template-http'
                if vthunder.partition_name != "shared":
                    if CONF.a10_global.use_shared_for_template_lookup:
                        template_key = utils.shared_template_modifier(template_key,
                                                                      template_http,
                                                                      device_templates)
                vport_templates[template_key] = template_http
            """
            if ha_conn_mirror is not None:
                ha_conn_mirror = None
                LOG.warning("'ha_conn_mirror' is not allowed for HTTP "
                            "or TERMINATED_HTTPS listeners.")
            """
        elif listener.protocol == 'tcp':
            template_tcp = CONF.listener.template_tcp
            if template_tcp and template_tcp.lower() != 'none':
                template_key = 'template-tcp'
                if vthunder.partition_name != "shared":
                    if CONF.a10_global.use_shared_for_template_lookup:
                        template_key = utils.shared_template_modifier(template_key,
                                                                      template_tcp,
                                                                      device_templates)
                vport_templates[template_key] = template_tcp

        template_policy = CONF.listener.template_policy
        if template_policy and template_policy.lower() != 'none':
            template_key = 'template-policy'
            if vthunder.partition_name != "shared":
                if CONF.a10_global.use_shared_for_template_lookup:
                    template_key = utils.shared_template_modifier(template_key,
                                                                  template_policy,
                                                                  device_templates)
            vport_templates[template_key] = template_policy

        vport_args = {}
        if flavor_data:
            virtual_port_flavor = flavor_data.get('virtual_port')
            if virtual_port_flavor:
                name_exprs = virtual_port_flavor.get('name_expressions')
                parsed_exprs = utils.parse_name_expressions(
                    listener.name, name_exprs)
                virtual_port_flavor.pop('name_expressions', None)
                virtual_port_flavor.update(parsed_exprs)
                vport_args = {'port': virtual_port_flavor}

            # use default nat-pool pool if pool is not specified
            if 'port' not in vport_args or 'pool' not in vport_args['port']:
                pool_flavor = flavor_data.get('nat_pool')
                if pool_flavor and 'pool_name' in pool_flavor:
                    pool_arg = {}
                    pool_arg['pool'] = pool_flavor['pool_name']
                    if 'port' in vport_args:
                        vport_args['port'].update(pool_arg)
                    else:
                        vport_args['port'] = pool_arg
        config_data.update(template_args)
        config_data.update(vport_args)

        set_method(loadbalancer.id,
                   listener.id,
                   listener.protocol,
                   listener.protocol_port,
                   listener.default_pool_id,
                   s_pers_name=s_pers, c_pers_name=c_pers,
                   virtual_port_templates=vport_templates,
                   **config_data)

        listener.protocol = listener.protocol.upper()


class ListenerCreate(ListenersParent, task.Task):
    """Task to create listener"""

    @axapi_client_decorator
    def execute(self, loadbalancer, listener, vthunder, flavor_data=None):
        try:
            self.set(self.axapi_client.slb.virtual_server.vport.create,
                     loadbalancer, listener, vthunder, flavor_data)
            LOG.debug("Successfully created listener: %s", listener.id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to create listener: %s", listener.id)
            raise e

    @axapi_client_decorator_for_revert
    def revert(self, loadbalancer, listener, vthunder, *args, **kwargs):
        LOG.warning("Reverting creation of listener: %s", listener.id)
        try:
            self.axapi_client.slb.virtual_server.vport.delete(
                loadbalancer.id, listener.id, listener.protocol,
                listener.protocol_port)
        except ConnectionError:
            LOG.exception(
                "Failed to connect A10 Thunder device: %s", vthunder.ip_address)
        except Exception as e:
            LOG.exception("Failed to revert creation of listener: %s due to %s",
                          listener.id, str(e))


class ListenerUpdate(ListenersParent, task.Task):
    """Task to update listener"""

    @axapi_client_decorator
    def execute(self, loadbalancer, listener, vthunder, flavor_data=None, update_dict=None):
        try:
            if listener:
                self.set(self.axapi_client.slb.virtual_server.vport.replace,
                         loadbalancer, listener, vthunder, flavor_data, update_dict)
                LOG.debug("Successfully updated listener: %s", listener.id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to update listener: %s", listener.id)
            raise e


class ListenerUpdateForPool(ListenersParent, task.Task):
    """Task to update listener while pool delete"""

    @axapi_client_decorator
    def execute(self, loadbalancer, listener, vthunder):
        try:
            if listener:
                c_pers, s_pers = utils.get_sess_pers_templates(listener.default_pool)
                listener.protocol = openstack_mappings.virtual_port_protocol(
                    self.axapi_client, listener.protocol).lower()
                self.axapi_client.slb.virtual_server.vport.update(
                    loadbalancer.id,
                    listener.id,
                    listener.protocol,
                    listener.protocol_port,
                    listener.default_pool_id,
                    s_pers_name=s_pers, c_pers_name=c_pers)
                LOG.debug("Successfully updated listener: %s", listener.id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to update listener: %s", listener.id)
            raise e


class ListenerDelete(ListenersParent, task.Task):
    """Task to delete the listener"""

    @axapi_client_decorator
    def execute(self, loadbalancer, listener, vthunder):
        listener.protocol = openstack_mappings.virtual_port_protocol(self.axapi_client,
                                                                     listener.protocol)
        try:
            self.axapi_client.slb.virtual_server.vport.delete(
                loadbalancer.id, listener.id, listener.protocol,
                listener.protocol_port)
            LOG.debug("Successfully deleted listener: %s", listener.id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to delete listener: %s", listener.id)
            raise e



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import imp
try:
    from unittest import mock
except ImportError:
    import mock

from octavia.common import data_models as o_data_models
from octavia.tests.common import constants as t_constants
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from a10_octavia.common.data_models import VThunder
import a10_octavia.controller.worker.tasks.virtual_server_tasks as task
from a10_octavia.tests.common import a10constants
from a10_octavia.tests.unit.base import BaseTaskTestCase

VTHUNDER = VThunder()
AMPHORA = o_data_models.Amphora(id=t_constants.MOCK_AMP_ID1)
VIP = o_data_models.Vip(ip_address="1.1.1.1")
LOADBALANCER = o_data_models.LoadBalancer(id=a10constants.MOCK_LOAD_BALANCER_ID,
                                          amphorae=[AMPHORA], vip=VIP)


class TestHandlerVirtualServerTasks(BaseTaskTestCase):

    def setUp(self):
        super(TestHandlerVirtualServerTasks, self).setUp()
        imp.reload(task)
        self.client_mock = mock.Mock()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))

    def tearDown(self):
        super(TestHandlerVirtualServerTasks, self).tearDown()
        self.conf.reset()

    def test_CreateVirtualServerTask_revert(self):
        mock_load_balancer = task.CreateVirtualServerTask()
        mock_load_balancer.axapi_client = self.client_mock
        mock_load_balancer.revert(LOADBALANCER, VTHUNDER)
        self.client_mock.slb.virtual_server.delete.assert_called_with(LOADBALANCER.id)

    @mock.patch('a10_octavia.controller.worker.tasks.utils.parse_name_expressions',
                mock.MagicMock())
    def test_CreateVirtualServerTask_execute_create_with_flavor(self):
        flavor = {"virtual_server": {"arp_disable": 1}}

        vthunder = copy.deepcopy(VTHUNDER)
        virtual_server_task = task.CreateVirtualServerTask()
        virtual_server_task.axapi_client = self.client_mock
        virtual_server_task.execute(LOADBALANCER, vthunder, flavor_data=flavor)
        args, kwargs = self.client_mock.slb.virtual_server.create.call_args
        self.assertEqual(kwargs['virtual_server']['arp_disable'], 1)

    @mock.patch('a10_octavia.controller.worker.tasks.utils.parse_name_expressions',
                mock.MagicMock())
    def test_CreateVirtualServerTask_execute_flavor_override_config(self):
        self.conf.config(group=a10constants.VIRTUAL_SERVER_CONFIG_SECTION,
                         arp_disable=False)
        flavor = {"virtual_server": {"arp_disable": 1}}

        vthunder = copy.deepcopy(VTHUNDER)
        virtual_server_task = task.CreateVirtualServerTask()
        virtual_server_task.axapi_client = self.client_mock
        virtual_server_task.execute(LOADBALANCER, vthunder, flavor_data=flavor)
        args, kwargs = self.client_mock.slb.virtual_server.create.call_args
        self.assertEqual(kwargs['virtual_server']['arp_disable'], 1)

    @mock.patch('a10_octavia.controller.worker.tasks.utils.parse_name_expressions')
    def test_CreateVirtualServerTask_execute_create_with_flavor_regex(self, mock_name_expr):
        mock_name_expr.return_value = {"arp_disable": 0}

        vthunder = copy.deepcopy(VTHUNDER)
        virtual_server_task = task.CreateVirtualServerTask()
        virtual_server_task.axapi_client = self.client_mock
        virtual_server_task.execute(LOADBALANCER, vthunder, flavor_data={})
        args, kwargs = self.client_mock.slb.virtual_server.create.call_args
        self.assertEqual(kwargs['arp_disable'], 0)

    @mock.patch('a10_octavia.controller.worker.tasks.utils.parse_name_expressions')
    def test_CreateVirtualServerTask_execute_name_expr_override_flavor(self, mock_name_expr):
        flavor = {"virtual_server": {"arp_disable": 1}}
        mock_name_expr.return_value = {"arp_disable": 0}

        vthunder = copy.deepcopy(VTHUNDER)
        virtual_server_task = task.CreateVirtualServerTask()
        virtual_server_task.axapi_client = self.client_mock
        virtual_server_task.execute(LOADBALANCER, vthunder, flavor_data=flavor)
        args, kwargs = self.client_mock.slb.virtual_server.create.call_args
        self.assertEqual(kwargs['arp_disable'], 0)



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from datetime import datetime
import json
import sqlalchemy
from sqlalchemy.orm.exc import NoResultFound

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
from taskflow import task
from taskflow.types import failure

from octavia.common import constants
from octavia.common import exceptions as octavia_exceptions
from octavia.db import api as db_apis
from octavia.db import repositories as repo

from a10_octavia.common import a10constants
from a10_octavia.common import exceptions
from a10_octavia.common import utils
from a10_octavia.controller.worker.tasks import utils as a10_task_utils
from a10_octavia.db import repositories as a10_repo

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class BaseDatabaseTask(task.Task):
    """Base task to load drivers common to the tasks."""

    def __init__(self, **kwargs):
        self.repos = repo.Repositories()
        self.vthunder_repo = a10_repo.VThunderRepository()
        self.vrid_repo = a10_repo.VRIDRepository()
        self.amphora_repo = repo.AmphoraRepository()
        self.member_repo = a10_repo.MemberRepository()
        self.loadbalancer_repo = a10_repo.LoadBalancerRepository()
        self.vip_repo = repo.VipRepository()
        self.listener_repo = repo.ListenerRepository()
        self.flavor_repo = repo.FlavorRepository()
        self.flavor_profile_repo = repo.FlavorProfileRepository()
        self.nat_pool_repo = a10_repo.NatPoolRepository()
        self.vrrp_set_repo = a10_repo.VrrpSetRepository()
        super(BaseDatabaseTask, self).__init__(**kwargs)


class GetVThunderTask(BaseDatabaseTask):
    """Test VThunder entry"""

    def execute(self, amphora):
        vthunder = self.vthunder_repo.get(db_apis.get_session(), id=123)
        return vthunder


class CreateVThunderEntry(BaseDatabaseTask):
    """ Create VThunder device entry in DB"""

    def execute(self, amphora, loadbalancer, role, status):
        vthunder_id = uuidutils.generate_uuid()

        username = CONF.vthunder.default_vthunder_username
        password = CONF.vthunder.default_vthunder_password
        axapi_version = CONF.vthunder.default_axapi_version
        topology = CONF.a10_controller_worker.loadbalancer_topology

        if role == constants.ROLE_MASTER:
            role = "MASTER"
        elif role == constants.ROLE_BACKUP:
            role = "BACKUP"
        else:
            role = "MASTER"

        self.vthunder_repo.create(
            db_apis.get_session(), vthunder_id=vthunder_id,
            amphora_id=amphora.id,
            device_name=vthunder_id, username=username,
            password=password, ip_address=amphora.lb_network_ip,
            undercloud=False, axapi_version=axapi_version,
            loadbalancer_id=loadbalancer.id,
            project_id=loadbalancer.project_id,
            compute_id=amphora.compute_id,
            topology=topology,
            role=role,
            last_udp_update=datetime.utcnow(),
            status=status,
            created_at=datetime.utcnow(),
            updated_at=datetime.utcnow())

        LOG.info("Successfully created vthunder entry in database.")

    def revert(self, amphora, loadbalancer, role, status=constants.ACTIVE,
               *args, **kwargs):
        try:
            self.vthunder_repo.delete(
                db_apis.get_session(), loadbalancer_id=loadbalancer.id)
        except NoResultFound:
            LOG.error(
                "Failed to delete vThunder entry for load balancer: %s",
                loadbalancer.id)


class CheckExistingThunderToProjectMappedEntries(BaseDatabaseTask):
    """ Check for all existing Thunder entries to ensure
        all belong to different projects, otherwise
        raise ConfigValueError
    """

    def execute(self, loadbalancer, vthunder_config, use_device_flavor=False):
        if use_device_flavor:
            return
        hierarchical_mt = vthunder_config.hierarchical_multitenancy
        if hierarchical_mt == 'enable' and CONF.a10_global.use_parent_partition:
            return
        vthunder_ids = self.vthunder_repo.get_vthunders_by_ip_address(
            db_apis.get_session(),
            ip_address=vthunder_config.ip_address)

        for vthunder_id in vthunder_ids:
            vthunder = self.vthunder_repo.get(
                db_apis.get_session(),
                id=vthunder_id)
            if vthunder is None:
                return

            existing_ip_addr_partition = '{}:{}'.format(
                vthunder.ip_address, vthunder.partition_name)
            config_ip_addr_partition = '{}:{}'.format(
                vthunder_config.ip_address, vthunder_config.partition_name)
            if existing_ip_addr_partition == config_ip_addr_partition:
                if loadbalancer.project_id not in (vthunder.project_id,
                                                   utils.get_parent_project(vthunder.project_id)):
                    raise exceptions.ProjectInUseByExistingThunderError(
                        config_ip_addr_partition, vthunder.project_id, loadbalancer.project_id)


class DeleteVThunderEntry(BaseDatabaseTask):
    """ Delete VThunder device entry in DB  """

    def execute(self, loadbalancer):
        try:
            self.vthunder_repo.delete(
                db_apis.get_session(), loadbalancer_id=loadbalancer.id)
        except NoResultFound:
            pass
        LOG.info("Successfully deleted vthunder entry in database.")


class GetVThunderByLoadBalancer(BaseDatabaseTask):
    """Get VThunder from db using LoadBalancer"""

    def execute(self, loadbalancer):
        loadbalancer_id = loadbalancer.id
        vthunder = self.vthunder_repo.get_vthunder_from_lb(
            db_apis.get_session(), loadbalancer_id)
        if vthunder is None:
            return None
        return vthunder


class GetBackupVThunderByLoadBalancer(BaseDatabaseTask):
    """ Get VThunder details from LoadBalancer"""

    def execute(self, loadbalancer, vthunder=None):
        loadbalancer_id = loadbalancer.id
        backup_vthunder = self.vthunder_repo.get_backup_vthunder_from_lb(
            db_apis.get_session(), loadbalancer_id)

        # VCS vMaster/vBlade may switched
        if vthunder is not None and backup_vthunder:
            if backup_vthunder.ip_address == vthunder.ip_address:
                backup_vthunder = self.vthunder_repo.get_vthunder_from_lb(
                    db_apis.get_session(), loadbalancer_id)
        return backup_vthunder
        LOG.info("Successfully fetched vThunder details for LB")


# class GetVThunderByLoadBalancerID(BaseDatabaseTask):
#     """ Get VThunder details from LoadBalancer ID """
#     def execute(self, loadbalancer_id):
#         vthunder = self.vthunder_repo.get_vthunder_from_lb(db_apis.get_session(), loadbalancer_id)
#         return vthunder
#         LOG.info("Successfully fetched vThunder details for LB")


class MapLoadbalancerToAmphora(BaseDatabaseTask):
    """Maps and assigns a load balancer to an amphora in the database."""

    def execute(self, loadbalancer, server_group_id=None):
        """Allocates an Amphora for the load balancer in the database.

        :param loadbalancer_id: The load balancer id to map to an amphora
        :returns: Amphora ID if one was allocated, None if it was
                  unable to allocate an Amphora
        """

        if server_group_id is not None:
            LOG.debug("Load balancer is using anti-affinity. Skipping spares "
                      "pool allocation.")
            return None

        vthunder = self.vthunder_repo.get_vthunder_by_project_id(
            db_apis.get_session(),
            loadbalancer.project_id)

        if vthunder is None:
            # Check for spare vthunder
            vthunder = self.vthunder_repo.get_spare_vthunder(
                db_apis.get_session())
            if vthunder is None:
                LOG.debug("No Amphora available for load balancer with id %s",
                          loadbalancer.id)
                return None

        return vthunder.id


class CreateRackVthunderEntry(BaseDatabaseTask):
    """ Create VThunder device entry in DB """

    def execute(self, loadbalancer, vthunder_config):
        hierarchical_mt = vthunder_config.hierarchical_multitenancy
        try:
            vthunder = self.vthunder_repo.create(
                db_apis.get_session(),
                vthunder_id=uuidutils.generate_uuid(),
                device_name=vthunder_config.device_name,
                username=vthunder_config.username,
                password=vthunder_config.password,
                ip_address=vthunder_config.ip_address,
                undercloud=vthunder_config.undercloud,
                loadbalancer_id=loadbalancer.id,
                project_id=vthunder_config.project_id,
                axapi_version=vthunder_config.axapi_version,
                topology="STANDALONE",
                role="MASTER",
                status="ACTIVE",
                last_udp_update=datetime.utcnow(),
                created_at=datetime.utcnow(),
                updated_at=datetime.utcnow(),
                partition_name=vthunder_config.partition_name,
                hierarchical_multitenancy=hierarchical_mt)
            LOG.info("Successfully created vthunder entry in database.")
            return vthunder
        except Exception as e:
            LOG.error(
                'Failed to create vThunder entry in db for load balancer: %s.',
                loadbalancer.id)
            raise e

    def revert(self, result, loadbalancer, vthunder_config, *args, **kwargs):
        if isinstance(result, failure.Failure):
            # This task's execute failed, so nothing needed to be done to
            # revert
            return

        LOG.warning(
            'Reverting create Rack VThunder in DB for load balancer: %s',
            loadbalancer.id)
        try:
            self.vthunder_repo.delete(
                db_apis.get_session(), loadbalancer_id=loadbalancer.id)
        except Exception:
            LOG.error(
                "Failed to delete vThunder entry for load balancer: %s",
                loadbalancer.id)


class CreateVThunderHealthEntry(BaseDatabaseTask):
    """ Create VThunder Health entry in DB """

    def execute(self, loadbalancer, vthunder_config):
        self.vthunder_repo.create(db_apis.get_session(),
                                  vthunder_id=uuidutils.generate_uuid(),
                                  device_name=vthunder_config.device_name,
                                  username=vthunder_config.username,
                                  password=vthunder_config.password,
                                  ip_address=vthunder_config.ip_address,
                                  undercloud=vthunder_config.undercloud,
                                  loadbalancer_id=loadbalancer.id,
                                  project_id=vthunder_config.project_id,
                                  axapi_version=vthunder_config.axapi_version,
                                  topology="STANDALONE",
                                  role="MASTER")
        LOG.info("Successfully created vthunder entry in database.")


class MarkVThunderStatusInDB(BaseDatabaseTask):

    def execute(self, vthunder, status):
        try:
            if vthunder:
                self.vthunder_repo.update(
                    db_apis.get_session(),
                    vthunder.id,
                    status=status,
                    updated_at=datetime.utcnow())
        except (sqlalchemy.orm.exc.NoResultFound,
                sqlalchemy.orm.exc.UnmappedInstanceError):
            LOG.debug('No existing amphora health record to mark busy '
                      'for amphora: %s, skipping.', vthunder.id)


class CreateSpareVThunderEntry(BaseDatabaseTask):

    def execute(self, amphora):
        vthunder_id = uuidutils.generate_uuid()

        username = CONF.vthunder.default_vthunder_username
        password = CONF.vthunder.default_vthunder_password
        axapi_version = CONF.vthunder.default_axapi_version
        vthunder = self.vthunder_repo.create(
            db_apis.get_session(), vthunder_id=vthunder_id,
            amphora_id=amphora.id,
            device_name=vthunder_id, username=username,
            password=password, ip_address=amphora.lb_network_ip,
            undercloud=False, axapi_version=axapi_version,
            loadbalancer_id=None,
            project_id=None,
            compute_id=amphora.compute_id,
            topology="SINGLE",
            role="MASTER",
            status="READY",
            created_at=datetime.utcnow(),
            last_udp_update=datetime.utcnow(),
            updated_at=datetime.utcnow())
        LOG.info("Successfully created vthunder entry in database.")
        return vthunder


class GetVRIDForLoadbalancerResource(BaseDatabaseTask):

    def execute(self, partition_project_list, vthunder):
        vrid_list = []
        if vthunder:
            owner = vthunder.ip_address + "_" + vthunder.partition_name
            if partition_project_list:
                try:
                    vrid_list = self.vrid_repo.get_vrid_from_owner(
                        db_apis.get_session(), owner=owner,
                        project_ids=partition_project_list)
                    return vrid_list
                except Exception as e:
                    LOG.exception(
                        "Failed to get VRID list for given owner  %s due to %s",
                        owner,
                        str(e))
                    raise e
        return vrid_list


class UpdateVRIDForLoadbalancerResource(BaseDatabaseTask):

    def __init__(self, *arg, **kwargs):
        self.vrid_created = []
        super(UpdateVRIDForLoadbalancerResource, self).__init__(*arg, **kwargs)

    def execute(self, lb_resource, vrid_list, vthunder):
        owner = vthunder.ip_address + "_" + vthunder.partition_name
        if not vrid_list:
            # delete all vrids from DB for the lB resource's project.
            try:
                self.vrid_repo.delete(db_apis.get_session(),
                                      owner=owner)
                LOG.debug("Successfully deleted DB vrid from owner %s", owner)

            except Exception as e:
                LOG.error(
                    "Failed to delete VRID data for owner %s due to %s",
                    owner,
                    str(e))
                raise e
            return
        for vrid in vrid_list:
            if self.vrid_repo.exists(db_apis.get_session(), vrid.id):
                try:
                    self.vrid_repo.update(
                        db_apis.get_session(),
                        vrid.id,
                        vrid_floating_ip=vrid.vrid_floating_ip,
                        vrid_port_id=vrid.vrid_port_id,
                        vrid=vrid.vrid,
                        subnet_id=vrid.subnet_id)
                    LOG.debug(
                        "Successfully updated DB vrid %s entry for loadbalancer resource %s",
                        vrid.id,
                        lb_resource.id)
                except Exception as e:
                    LOG.error(
                        "Failed to update VRID data for VRID FIP %s due to %s",
                        vrid.vrid_floating_ip,
                        str(e))
                    raise e

            else:
                try:
                    new_vrid = self.vrid_repo.create(
                        db_apis.get_session(),
                        id=vrid.id,
                        owner=owner,
                        vrid_floating_ip=vrid.vrid_floating_ip,
                        vrid_port_id=vrid.vrid_port_id,
                        vrid=vrid.vrid,
                        subnet_id=vrid.subnet_id)
                    self.vrid_created.append(new_vrid)
                except Exception as e:
                    LOG.error(
                        "Failed to create VRID data for VRID FIP %s due to %s",
                        vrid.vrid_floating_ip,
                        str(e))
                    raise e

    def revert(self, *args, **kwargs):
        for vrid in self.vrid_created:
            try:
                self.vrid_repo.delete(db_apis.get_session(), id=vrid.id)
            except Exception as e:
                LOG.error("Failed to update VRID DB entry %s due to %s",
                          str(vrid.vrid_floating_ip), str(e))


class GetLoadbalancersInProjectBySubnet(BaseDatabaseTask):

    def execute(self, subnet, partition_project_list):
        if partition_project_list:
            try:
                return self.loadbalancer_repo.get_lbs_by_subnet(
                    db_apis.get_session(),
                    project_ids=partition_project_list, subnet_id=subnet.id)
            except Exception as e:
                LOG.exception("Failed to get LBs for subnet %s due to %s ",
                              subnet.id, str(e))
                raise e
        return 0


class CountLoadbalancersInProjectBySubnet(BaseDatabaseTask):

    def execute(self, subnet, partition_project_list, use_device_flavor):
        if not subnet:
            return 0
        if not use_device_flavor and partition_project_list:
            try:
                return self.loadbalancer_repo.get_lb_count_by_subnet(
                    db_apis.get_session(),
                    project_ids=partition_project_list, subnet_id=subnet.id)
            except Exception as e:
                LOG.exception("Failed to get LB count for subnet %s due to %s ",
                              subnet.id, str(e))
                raise e
        return 0


class CountMembersInProjectBySubnet(BaseDatabaseTask):

    def execute(self, subnet, partition_project_list):
        if not subnet:
            return 0
        if partition_project_list:
            try:
                return self.member_repo.get_member_count_by_subnet(
                    db_apis.get_session(),
                    project_ids=partition_project_list, subnet_id=subnet.id)
            except Exception as e:
                LOG.exception(
                    "Failed to get LB member count for subnet %s due to %s",
                    subnet.id, str(e))
                raise e
        return 0


class DeleteVRIDEntry(BaseDatabaseTask):
    def execute(self, vrid, delete_vrid):
        if vrid and delete_vrid:
            try:
                self.vrid_repo.delete(db_apis.get_session(), id=vrid.id)
            except Exception as e:
                LOG.exception(
                    "Failed to delete VRID entry from vrid table: %s", str(e))
                raise e


class DeleteMultiVRIDEntry(BaseDatabaseTask):
    def execute(self, vrid_list):
        if vrid_list:
            try:
                self.vrid_repo.delete_batch(
                    db_apis.get_session(), ids=[
                        vrid.id for vrid in vrid_list])
            except Exception as e:
                LOG.exception(
                    "Failed to delete VRID entry from vrid table: %s", str(e))
                raise e


class CheckVLANCanBeDeletedParent(object):
    """ Checks all vip and member subnet_ids for project ID"""

    def is_vlan_deletable(self, project_id, subnet_id, is_vip):
        if project_id not in CONF.rack_vthunder.devices:
            return False

        subnet_usage_count = 0

        lbs = db_apis.get_session().query(self.loadbalancer_repo.model_class).filter(
            self.loadbalancer_repo.model_class.project_id == project_id).all()
        for lb in lbs:
            vips = db_apis.get_session().query(
                self.vip_repo.model_class).filter(
                self.vip_repo.model_class.load_balancer_id == lb.id).all()
            for vip in vips:
                if vip.subnet_id == subnet_id:
                    subnet_usage_count += 1

        members = db_apis.get_session().query(
            self.member_repo.model_class).filter(
            self.member_repo.model_class.project_id == project_id).all()
        for member in members:
            if member.subnet_id == subnet_id:
                subnet_usage_count += 1

        if is_vip and subnet_usage_count == 1:
            return True
        elif subnet_usage_count == 0:
            return True

        return False


class CheckVipVLANCanBeDeleted(CheckVLANCanBeDeletedParent, BaseDatabaseTask):
    default_provides = a10constants.DELETE_VLAN

    def execute(self, loadbalancer):
        return self.is_vlan_deletable(loadbalancer.project_id,
                                      loadbalancer.vip.subnet_id,
                                      True)


class CheckMemberVLANCanBeDeleted(
        CheckVLANCanBeDeletedParent,
        BaseDatabaseTask):
    default_provides = a10constants.DELETE_VLAN

    def execute(self, member):
        return self.is_vlan_deletable(
            member.project_id, member.subnet_id, False)


class MarkLBAndListenerActiveInDB(BaseDatabaseTask):
    """Mark the load balancer and specified listener active in the DB"""

    def execute(self, loadbalancer, listener):
        """Mark the load balancer and listener as active in DB"""

        self.loadbalancer_repo.update(db_apis.get_session(),
                                      loadbalancer.id,
                                      provisioning_status=constants.ACTIVE)
        self.listener_repo.prov_status_active_if_not_error(
            db_apis.get_session(), listener.id)

    def revert(self, loadbalancer, listener, *args, **kwargs):
        """Mark the load balancer and listener in error state"""
        try:
            self.loadbalancer_repo.update(db_apis.get_session(),
                                          id=loadbalancer.id,
                                          provisioning_status=constants.ERROR)
        except Exception as e:
            LOG.error("Failed to update load balancer %(lb) "
                      "provisioning status to ERROR due to: "
                      "%(except)s", {'lb': loadbalancer.id, 'except': e})
        try:
            self.listener_repo.update(db_apis.get_session(),
                                      id=listener.id,
                                      provisioning_status=constants.ERROR)
        except Exception as e:
            LOG.error("Failed to update listener %(list) "
                      "provisioning status to ERROR due to: "
                      "%(except)s", {'list': listener.id, 'except': e})


# class GetVRIDForLBResourceSubnet(BaseDatabaseTask):
#     def execute(self, lb_resource, subnet):
#         project_id = lb_resource.project_id
#         vrid = self.vrid_repo.get_vrid_for_subnet(
#             db_apis.get_session(), project_id=project_id, subnet_id=subnet.id)
#         return vrid


class CountMembersWithIP(BaseDatabaseTask):
    def execute(self, member):
        try:
            return self.member_repo.get_member_count_by_ip_address(
                db_apis.get_session(), member.ip_address, member.project_id)
        except Exception as e:
            LOG.exception(
                "Failed to get count of members with given IP for a pool: %s",
                str(e))
            raise e


class CountMembersWithIPPortProtocol(BaseDatabaseTask):
    def execute(self, member, pool):
        try:
            return self.member_repo.get_member_count_by_ip_address_port_protocol(
                db_apis.get_session(), member.ip_address, member.project_id,
                member.protocol_port, pool.protocol)
        except Exception as e:
            LOG.exception(
                "Failed to get count of members with given IP fnd port for a pool: %s",
                str(e))
            raise e


class PoolCountforIP(BaseDatabaseTask):
    def execute(self, member, use_device_flavor, pools):
        try:
            if use_device_flavor:
                return self.member_repo.get_pool_count_by_ip_on_thunder(
                    db_apis.get_session(), member.ip_address, pools)
            else:
                return self.member_repo.get_pool_count_by_ip(
                    db_apis.get_session(), member.ip_address, member.project_id)
        except Exception as e:
            LOG.exception(
                "Failed to get pool count with same IP address: %s",
                str(e))
            raise e


class GetSubnetForDeletionInPool(BaseDatabaseTask):

    def execute(self, member_list, partition_project_list, use_device_flavor, pools):
        """

        :param member_list: Receives the list of members, under specific pool.
        :param partition_project_list: list of projects under partition.
        :return: returns the list subnet those used in a pool's members only.
        Description: Iterates over the member list, to check if member's  subnet is
        anywhere else, along with pool count is also takes LB into account.
        """
        try:
            subnet_list = []
            member_subnet = []
            for member in member_list:
                if member.subnet_id not in member_subnet:
                    if use_device_flavor:
                        pool_count_subnet = self.member_repo.get_pool_count_subnet_on_thunder(
                            db_apis.get_session(), pools, member.subnet_id)
                    else:
                        pool_count_subnet = self.member_repo.get_pool_count_subnet(
                            db_apis.get_session(), partition_project_list, member.subnet_id)
                    lb_count_subnet = self.loadbalancer_repo.get_lb_count_by_subnet(
                        db_apis.get_session(), partition_project_list, member.subnet_id)
                    if pool_count_subnet <= 1 and lb_count_subnet == 0:
                        subnet_list.append(member.subnet_id)
                    member_subnet.append(member.subnet_id)
            return subnet_list
        except Exception as e:
            LOG.exception("Failed to get subnet list for members: %s", str(e))
            raise e


class GetChildProjectsOfParentPartition(BaseDatabaseTask):

    def execute(self, lb_resource, vthunder):
        if vthunder:
            try:
                partition_project_list = self.vthunder_repo.get_project_list_using_partition(
                    db_apis.get_session(),
                    partition_name=vthunder.partition_name,
                    ip_address=vthunder.ip_address)
                return partition_project_list
            except Exception as e:
                LOG.exception(
                    "Failed to fetch list of projects, if multi-tenancy and use parent partition "
                    "is enabled due to %s", str(e))
                raise e


class GetFlavorData(BaseDatabaseTask):

    def _format_keys(self, flavor_data):
        if type(flavor_data) is list:
            item_list = []
            for item in flavor_data:
                item_list.append(self._format_keys(item))
            return item_list
        elif type(flavor_data) is dict:
            item_dict = {}
            for k, v in flavor_data.items():
                item_dict[k.replace('-', '_')] = self._format_keys(v)
            return item_dict
        else:
            return flavor_data

    def execute(self, lb_resource):
        flavor_id = a10_task_utils.attribute_search(lb_resource, 'flavor_id')
        if not flavor_id:
            flavor_id = CONF.a10_global.default_flavor_id
        if flavor_id:
            flavor = self.flavor_repo.get(db_apis.get_session(), id=flavor_id)
            if not flavor and lb_resource.provisioning_status != "PENDING_DELETE":
                raise exceptions.FlavorNotFound(flavor_id)
            if flavor and flavor.flavor_profile_id:
                flavor_profile = self.flavor_profile_repo.get(
                    db_apis.get_session(),
                    id=flavor.flavor_profile_id)
                flavor_data = json.loads(flavor_profile.flavor_data)
                return self._format_keys(flavor_data)


class CheckForL2DSRFlavor(BaseDatabaseTask):

    def execute(self, lb_resource):
        for lb in lb_resource:
            flavor_id = a10_task_utils.attribute_search(lb, 'flavor_id')
            if flavor_id:
                flavor = self.flavor_repo.get(db_apis.get_session(), id=flavor_id)
                if flavor and flavor.flavor_profile_id:
                    flavor_profile = self.flavor_profile_repo.get(
                        db_apis.get_session(),
                        id=flavor.flavor_profile_id)
                    flavor_data = json.loads(flavor_profile.flavor_data)
                    deployment = flavor_data.get('deployment')
                    if deployment and 'dsr_type' in deployment and \
                            deployment['dsr_type'] == "l2dsr_transparent":
                        return True
        return False


class GetNatPoolEntry(BaseDatabaseTask):

    def execute(self, member, nat_flavor=None):
        if nat_flavor and 'pool_name' in nat_flavor:
            try:
                return self.nat_pool_repo.get(db_apis.get_session(), name=nat_flavor['pool_name'],
                                              subnet_id=member.subnet_id)
            except Exception as e:
                LOG.exception("Failed to fetch subnet %s NAT pool %s entry from database: %s",
                              nat_flavor['pool_name'], member.subnet_id, str(e))
                raise e


class UpdateNatPoolDB(BaseDatabaseTask):

    def execute(self, member, nat_flavor=None, nat_pool=None, subnet_port=None):
        if nat_flavor is None:
            return

        if nat_pool is None:
            if subnet_port is None:
                # NAT pool addresses are not in member subnet. a10-octavia allows it
                # but not able to reerve ip for it. So, no database entry is needed.
                return

            try:
                id = uuidutils.generate_uuid()
                pool_name = nat_flavor.get('pool_name')
                subnet_id = member.subnet_id
                start_address = nat_flavor.get('start_address')
                end_address = nat_flavor.get('end_address')
                port_id = subnet_port.id
                self.nat_pool_repo.create(
                    db_apis.get_session(), id=id, name=pool_name, subnet_id=subnet_id,
                    start_address=start_address, end_address=end_address,
                    member_ref_count=1, port_id=port_id)
                LOG.info("Successfully created nat pool entry in database.")
            except Exception as e:
                LOG.exception("Failed to create subnet %s NAT pool %s entry to database: %s",
                              subnet_id, pool_name, str(e))
                raise e
        else:
            try:
                ref = nat_pool.member_ref_count + 1
                self.nat_pool_repo.update(
                    db_apis.get_session(), nat_pool.id, member_ref_count=ref)
            except Exception as e:
                LOG.exception("Failed to update NAT pool entry %s to database: %s",
                              nat_pool.id, str(e))
                raise e


class DeleteNatPoolEntry(BaseDatabaseTask):

    def execute(self, nat_pool=None):
        if nat_pool is None:
            return

        if nat_pool.member_ref_count > 1:
            try:
                ref = nat_pool.member_ref_count - 1
                self.nat_pool_repo.update(
                    db_apis.get_session(), nat_pool.id, member_ref_count=ref)
            except Exception as e:
                LOG.exception("Failed to update NAT pool entry %s to database: %s",
                              nat_pool.id, str(e))
                raise e
        else:
            try:
                self.nat_pool_repo.delete(db_apis.get_session(), id=nat_pool.id)
                LOG.info("Successfully deleted nat pool entry in database.")
            except Exception as e:
                LOG.exception("Failed to delete NAT pool entry %s to database: %s",
                              nat_pool.id, str(e))
                raise e


class CountLoadbalancersWithFlavor(BaseDatabaseTask):

    def execute(self, loadbalancer, vthunder):
        try:
            project_list = []
            if vthunder and vthunder.hierarchical_multitenancy == 'enable':
                project_list = self.vthunder_repo.get_project_list_using_partition(
                    db_apis.get_session(),
                    partition_name=vthunder.partition_name,
                    ip_address=vthunder.ip_address)
            else:
                project_list = [loadbalancer.project_id]

            return self.loadbalancer_repo.get_lb_count_by_flavor(
                db_apis.get_session(), project_list, loadbalancer.flavor_id)
        except Exception as e:
            LOG.exception("Failed to get LB count for flavor %s due to %s ",
                          loadbalancer.flavor_id, str(e))
            raise e
        return 0


class SetThunderUpdatedAt(BaseDatabaseTask):

    def execute(self, vthunder):
        try:
            if vthunder:
                LOG.debug("Updated the updated_at field for thunder : {}:{}"
                          .format(vthunder.ip_address, vthunder.partition_name))
                self.vthunder_repo.update(
                    db_apis.get_session(),
                    vthunder.id,
                    updated_at=datetime.utcnow())
        except Exception as e:
            LOG.exception('Failed to set updated_at field for thunder due to: {}'
                          ', skipping.'.format(str(e)))


class SetThunderLastWriteMem(BaseDatabaseTask):

    def execute(self, vthunder, write_mem_shared=False, write_mem_private=False):
        if write_mem_shared is False or write_mem_private is False:
            return
        try:
            if vthunder:
                LOG.debug("Updated the last_write_mem field for thunder : {}:{}"
                          .format(vthunder.ip_address, vthunder.partition_name))
                self.vthunder_repo.update_last_write_mem(
                    db_apis.get_session(),
                    vthunder.ip_address,
                    vthunder.partition_name,
                    last_write_mem=datetime.utcnow())
        except Exception as e:
            LOG.exception('Failed to set last_write_mem field for thunder due to: {}'
                          ', skipping.'.format(str(e)))


class GetActiveLoadBalancersByThunder(BaseDatabaseTask):

    def execute(self, vthunder):
        try:
            if vthunder:
                loadbalancers_list = self.loadbalancer_repo.get_active_lbs_by_thunder(
                    db_apis.get_session(),
                    vthunder)
                return loadbalancers_list
        except Exception as e:
            LOG.exception('Failed to get active Loadbalancers related to thunder '
                          'due to: {}'.format(str(e)))


class MarkLoadBalancersPendingUpdateInDB(BaseDatabaseTask):

    def execute(self, loadbalancers_list):
        try:
            for lb in loadbalancers_list:
                if lb.provisioning_status == constants.ERROR:
                    continue
                self.loadbalancer_repo.update(
                    db_apis.get_session(),
                    lb.id,
                    provisioning_status='PENDING_UPDATE')
        except Exception as e:
            LOG.exception('Failed to set Loadbalancers to PENDING_UPDATE due to '
                          ': {}'.format(str(e)))


class MarkLoadBalancersActiveInDB(BaseDatabaseTask):

    def execute(self, loadbalancers_list):
        try:
            for lb in loadbalancers_list:
                if lb.provisioning_status == constants.ERROR:
                    continue
                self.loadbalancer_repo.update(
                    db_apis.get_session(),
                    lb.id,
                    provisioning_status='ACTIVE')
        except Exception as e:
            LOG.exception('Failed to set Loadbalancers to ACTIVE due to '
                          ': {}'.format(str(e)))


class GetLoadBalancerListByProjectID(BaseDatabaseTask):

    def execute(self, vthunder):
        try:
            if vthunder:
                loadbalancers_list = self.loadbalancer_repo.get_lbs_by_project_id(
                    db_apis.get_session(),
                    vthunder.project_id)
                return loadbalancers_list
        except Exception as e:
            LOG.exception('Failed to get active Loadbalancers related to vthunder '
                          'due to: {}'.format(str(e)))


class AddProjectSetIdDB(BaseDatabaseTask):

    def execute(self, loadbalancer):
        set_id = 0
        if len(CONF.a10_controller_worker.amp_boot_network_list) < 1:
            raise exceptions.NoVrrpMgmtSubnet()
        mgmt_subnet = CONF.a10_controller_worker.amp_boot_network_list[0]

        try:
            project_id = loadbalancer.project_id
            entry = self.vrrp_set_repo.get(db_apis.get_session(), mgmt_subnet=mgmt_subnet,
                                           project_id=project_id)
            if entry is None:
                # amp_boot_network_list may changed but thunder already created with old subnet
                entry = self.vrrp_set_repo.get(db_apis.get_session(), project_id=project_id)

            if entry is not None:
                return entry.set_id

            subnet_set_ids = self.vrrp_set_repo.get_subnet_set_ids(db_apis.get_session(),
                                                                   mgmt_subnet)
            # ACOS allows set_id from 1-15
            while set_id < 15:
                set_id = set_id + 1
                if set_id not in subnet_set_ids:
                    break
            else:
                raise exceptions.NoFreeSetId()

            self.vrrp_set_repo.create(
                db_apis.get_session(), mgmt_subnet=mgmt_subnet, project_id=project_id,
                set_id=set_id)
            LOG.info("Successfully created set_id %d for project %s in database.",
                     set_id, project_id)
        except Exception as e:
            LOG.exception('Failed to get VRRP-A set_id for project due to :{}'.format(str(e)))

        return set_id


class DeleteProjectSetIdDB(BaseDatabaseTask):

    def execute(self, loadbalancer):
        try:
            # delete project set_id for all mgmt_subnet
            self.vrrp_set_repo.delete(db_apis.get_session(), project_id=loadbalancer.project_id)
        except Exception as e:
            LOG.exception('Failed to delete VRRP-A set_id for project due to :{}'.format(str(e)))


class GetLoadBalancerListForDeletion(BaseDatabaseTask):

    def execute(self, vthunder, loadbalancer):
        try:
            if vthunder:
                loadbalancers_list = self.loadbalancer_repo.get_all_other_lbs_in_project(
                    db_apis.get_session(),
                    vthunder.project_id,
                    loadbalancer.id)
                return loadbalancers_list
        except Exception as e:
            LOG.exception('Failed to get active Loadbalancers related to vthunder '
                          'due to: {}'.format(str(e)))


class CheckExistingVthunderTopology(BaseDatabaseTask):
    """This task only meant to use with vthunder flow[amphora]"""

    def execute(self, loadbalancer, topology):
        vthunder = self.vthunder_repo.get_vthunder_with_different_topology(
            db_apis.get_session(),
            loadbalancer.project_id,
            topology)

        if vthunder is not None:
            LOG.error('Other loadbalancer exists in vthunder with: %s topology',
                      vthunder.topology)
            msg = ('vthunder has other loadbalancer with {0} ' +
                   'topology').format(vthunder.topology)
            raise octavia_exceptions.InvalidTopology(topology=msg)


class ValidateComputeForProject(BaseDatabaseTask):
    """Validate compute_id got for the project"""

    def execute(self, loadbalancer, role):
        vthunder = None
        vthunder_ids = self.vthunder_repo.get_vthunders_by_project_id_and_role(
            db_apis.get_session(), loadbalancer.project_id, role)
        for vthunder_id in vthunder_ids:
            vthunder = self.vthunder_repo.get(
                db_apis.get_session(),
                id=vthunder_id)
            lb = self.loadbalancer_repo.get_lb_excluding_deleted(
                db_apis.get_session(), vthunder.loadbalancer_id)
            if lb:
                amphora = self.amphora_repo.get(db_apis.get_session(), load_balancer_id=lb.id)
                if not amphora:
                    """
                        We don't support create loadbalancer in Thunder device and vthunder in
                    same project now. So, this should caused by some error configuration.
                    """

                    LOG.error("Already use hardware thunder to create loadbalancer for"
                              " this project. Can't create vthunder for the project.")
                    raise exceptions.ProjectDeviceNotFound()
                if amphora.compute_id == vthunder.compute_id:
                    LOG.debug("Successfully validated comput_id %s for the project %s",
                              vthunder.compute_id, vthunder.project_id)
                    break

        if vthunder is None:
            return None
        return vthunder.compute_id


class CountLoadbalancersOnThunderBySubnet(BaseDatabaseTask):

    def execute(self, vthunder, subnet, use_device_flavor):
        if vthunder and use_device_flavor:
            try:
                count = 0
                vthunder_ids = self.vthunder_repo.get_rack_vthunders_by_ip_address(
                    db_apis.get_session(),
                    ip_address=vthunder.ip_address)
                for vthunder_id in vthunder_ids:
                    vth = self.vthunder_repo.get(
                        db_apis.get_session(),
                        id=vthunder_id)

                    if vthunder.partition_name == vth.partition_name:
                        lb = self.loadbalancer_repo.get_lbs_on_thunder_by_subnet(
                            db_apis.get_session(),
                            vth.loadbalancer_id,
                            subnet_id=subnet.id)
                        if lb:
                            count = count + 1
                return count
            except Exception as e:
                LOG.exception("Failed to get LB count on thunder for subnet %s due to %s ",
                              subnet.id, str(e))
                raise e
        return 0


class CountMembersOnThunderBySubnet(BaseDatabaseTask):

    def execute(self, subnet, use_device_flavor, members):
        if use_device_flavor:
            try:
                count = 0
                if members:
                    for mem in range(len(members)):
                        member = self.member_repo.get_members_on_thunder_by_subnet(
                            db_apis.get_session(),
                            ip_address=members[mem], subnet_id=subnet.id)
                        if member:
                            count = count + 1
                return count
            except Exception as e:
                LOG.exception("Failed to get member count on thunder for subnet %s due to %s ",
                              subnet.id, str(e))
                raise e
        return 0



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import logging
import re

from oslo_config import cfg

from a10_octavia.common.data_models import Certificate

from acos_client.v30 import responses as acos_resp
from requests import exceptions as req_exceptions

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def get_cert_data(barbican_client, listener):
    cert_data = Certificate()
    cert_ref = listener.tls_certificate_id
    cert_container = barbican_client.containers.get(container_ref=cert_ref)
    cert_data = Certificate(cert_filename=cert_container.certificate.name,
                            key_filename=cert_container.private_key.name,
                            cert_content=cert_container.certificate.payload,
                            key_content=cert_container.private_key.payload,
                            key_pass=cert_container.private_key_passphrase,
                            template_name=listener.id)
    return cert_data


def get_sess_pers_templates(pool):
    c_pers, s_pers, sp = None, None, None
    if pool and pool.session_persistence:
        sp = pool.session_persistence
        if sp.type == 'HTTP_COOKIE' or sp.type == 'APP_COOKIE':
            c_pers = pool.id
        elif sp.type == 'SOURCE_IP':
            s_pers = pool.id
    return c_pers, s_pers


def meta(lbaas_obj, key, default):
    if isinstance(lbaas_obj, dict):
        meta = lbaas_obj.get('a10_meta', '{}')
    elif hasattr(lbaas_obj, 'a10_meta'):
        meta = lbaas_obj.a10_meta
    else:
        return default
    try:
        meta_json = json.loads(meta)
    except Exception:
        return default
    return meta_json.get(key, default)


def shared_template_modifier(template_type, template_name, device_templates):
    resource_type = template_type.split('-', 1)[1]
    resource_list_key = "{0}-list".format(resource_type)
    if resource_list_key in device_templates:
        for device_template in device_templates['template'][resource_list_key]:
            device_template_name = device_template[resource_type].get("name")
            if template_name == device_template_name:
                break
            template_type = "{0}-shared".format(template_type)
    else:
        template_type = "{0}-shared".format(template_type)
    return template_type


def parse_name_expressions(name, name_expressions):
    flavor_data = {}
    if name and name_expressions:
        for expression in name_expressions:
            if 'regex' in expression:
                if re.search(expression['regex'], name):
                    flavor_data.update(expression['json'])
    return flavor_data


def dash_to_underscore(my_dict):
    if type(my_dict) is list:
        item_list = []
        for item in my_dict:
            item_list.append(dash_to_underscore(item))
        return item_list
    elif type(my_dict) is dict:
        item_dict = {}
        for k, v in my_dict.items():
            item_dict[k.replace('-', '_')] = dash_to_underscore(v)
        return item_dict
    else:
        return my_dict


def attribute_search(lb_resource, attr_name):
    """This helper method will recursively walk up the slb tree
    starting from the provided lb_resource and find an attribute
    with the provided name. Though objects like pool refrence a
    listener and loadbalancer, the following discrete
    search orders are used:

    1: member -> pool -> listener -> loadbalancer
    2: healthmonitor -> pool -> listener -> loadbalancer

    :param lb_resource: An slb data model
    :param obj_type: String name of an slb object
    (ie 'loadbalancer', 'pool')

    :return: Returns the requested attribute value or none
    """
    if hasattr(lb_resource, attr_name):
        return getattr(lb_resource, attr_name)
    elif hasattr(lb_resource, 'pool'):
        return attribute_search(lb_resource.pool, attr_name)
    elif hasattr(lb_resource, 'listener'):
        return attribute_search(lb_resource.listener, attr_name)
    elif hasattr(lb_resource, 'load_balancer'):
        return attribute_search(lb_resource.load_balancer, attr_name)
    return None


def thunder_busy_exceptions():
    return (acos_resp.axapi_retry_exceptions(), req_exceptions.ReadTimeout)



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.]

from octavia.tests.unit import base

from a10_octavia.controller.worker.tasks import utils

NAME_EXPRESSIONS = [
    {
        "regex": "vport1",
        "json": {"support_http2": 1}
    },
    {
        "regex": "vport2",
        "json": {"support_http2": 0}
    }
]

DASH_FLAVOR = {
    "virtual-server": {
        "port-number": 80,
        "name-expressions": [
            {
                "regex": "vip1-test",
                "json": {
                    "arp-disable": 0,
                    "extended-stats": 0,
                }
            },
            {
                "regex": "vip2-test",
                "json": {
                    "arp-disable": 1,
                    "extended-stats": 1,
                }
            }
        ]
    },
    "virtual-port": {
        "support-http2": 0
    },
}

UNDERSCORE_FLAVOR = {
    "virtual_server": {
        "port_number": 80,
        "name_expressions": [
            {
                "regex": "vip1-test",
                "json": {
                    "arp_disable": 0,
                    "extended_stats": 0,
                }
            },
            {
                "regex": "vip2-test",
                "json": {
                    "arp_disable": 1,
                    "extended_stats": 1,
                }
            }
        ]
    },
    "virtual_port": {
        "support_http2": 0
    },
}


class TestUtils(base.TestCase):

    def _get_device_templates(self, template_type, template_name):
        device_templates = {"template": {
            "{}-list".format(template_type): [{
                template_type: {"name": template_name}
            }]
        }
        }
        return device_templates

    def test_shared_template_modifier_resource_not_found(self):
        device_templates = self._get_device_templates('http', 'my_http_temp')
        template_type = utils.shared_template_modifier('template-tcp',
                                                       'my_tcp_temp',
                                                       device_templates)
        self.assertEqual('template-tcp-shared', template_type)

    def test_shared_template_modifier_resource_found_name_found(self):
        device_templates = self._get_device_templates('tcp', 'my_tcp_temp')
        template_type = utils.shared_template_modifier('template-tcp',
                                                       'my_tcp_temp',
                                                       device_templates)
        self.assertEqual('template-tcp-shared', template_type)

    def test_shared_template_modifier_resource_found_name_not_found(self):
        device_templates = self._get_device_templates('tcp', 'my_secondary_tcp')
        template_type = utils.shared_template_modifier('template-tcp',
                                                       'my_tcp_temp',
                                                       device_templates)
        self.assertEqual('template-tcp-shared', template_type)

    def test_parse_name_expressions(self):
        expect_result = {"support_http2": 1}
        obj_flavor = utils.parse_name_expressions("vport1", NAME_EXPRESSIONS)
        self.assertEqual(expect_result, obj_flavor)

    def test_dash_to_underscore(self):
        obj_flavor = utils.dash_to_underscore(DASH_FLAVOR)
        self.assertEqual(UNDERSCORE_FLAVOR, obj_flavor)

    def test_parse_name_expressions_with_no_name_specified(self):
        expect_result = {}
        obj_flavor = utils.parse_name_expressions(None, NAME_EXPRESSIONS)
        self.assertEqual(expect_result, obj_flavor)



# Copyright 2019 A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import socket

from oslo_config import cfg
from oslo_log import log as logging

from octavia.common import exceptions
from octavia.db import api as db_api

from a10_octavia.db import repositories as a10repo

UDP_MAX_SIZE = 64 * 1024
CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class VThunderUDPStatusGetter(object):
    """This class defines methods that will gather heatbeats.
    """

    def __init__(self):
        self.key = CONF.a10_health_manager.heartbeat_key
        self.ip = CONF.a10_health_manager.bind_ip
        self.port = CONF.a10_health_manager.bind_port
        self.sockaddr = None
        LOG.info('attempting to listen on %(ip)s port %(port)s',
                 {'ip': self.ip, 'port': self.port})
        self.sock = None
        self.update(self.key, self.ip, self.port)
        self.vthunder_repo = a10repo.VThunderRepository()

    def update(self, key, ip, port):
        """Update the running config for the udp socket server

        :param key: The hmac key used to verify the UDP packets. String
        :param ip: The ip address the UDP server will read from
        :param port: The port the UDP server will read from
        :return: None
        """
        self.key = key
        for addrinfo in socket.getaddrinfo(ip, port, 0, socket.SOCK_DGRAM):
            ai_family = addrinfo[0]
            self.sockaddr = addrinfo[4]
            if self.sock is not None:
                self.sock.close()
            self.sock = socket.socket(ai_family, socket.SOCK_DGRAM)
            self.sock.settimeout(1)
            self.sock.bind(self.sockaddr)
            if CONF.a10_health_manager.sock_rlimit > 0:
                rlimit = CONF.a10_health_manager.sock_rlimit
                LOG.info("setting sock rlimit to %s", rlimit)
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF,
                                     rlimit)
            break  # just used the first addr getaddrinfo finds
        if self.sock is None:
            raise exceptions.NetworkConfig("Unable to find suitable socket")

    def check(self):
        """Wait and obtain the source address of UDP packet and updates its time in
           vThunder repositories.
        """
        try:
            data, srcaddr = self.sock.recvfrom(UDP_MAX_SIZE)
            ip, port = srcaddr
            LOG.warning('Received packet from %s', ip)
            # get record id of first vThunder from srcaddr
            record_id = self.vthunder_repo.get_vthunder_from_src_addr(db_api.get_session(), ip)

            if record_id:
                last_udp_update = datetime.datetime.utcnow()
                self.vthunder_repo.update(db_api.get_session(), record_id,
                                          last_udp_update=last_udp_update)
        except socket.timeout:
            # Pass here as this is an expected cycling of the listen socket
            pass
        except exceptions.InvalidHMACException:
            # Pass here as the packet was dropped and logged already
            pass
        except Exception as ex:
            LOG.warning('Health Manager experienced an exception processing a'
                        'heartbeat packet. Ignoring this packet. '
                        'Exception: %s', ex)



"""adding HA features

Revision ID: ed9c523cf8ee
Revises: 8151f374ba74
Create Date: 2019-09-03 12:19:44.353070

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "ed9c523cf8ee"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'vthunders',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('vthunder_id', sa.String(36), nullable=False),
        sa.Column('amphora_id', sa.String(36), nullable=True),
        sa.Column('device_name', sa.String(1024), nullable=False),
        sa.Column('ip_address', sa.String(64), nullable=False),
        sa.Column('username', sa.String(1024), nullable=False),
        sa.Column('password', sa.String(50), nullable=False),
        sa.Column('axapi_version', sa.Integer, default=30, nullable=False),
        sa.Column('undercloud', sa.Boolean(), default=False, nullable=False),
        sa.Column('loadbalancer_id', sa.String(36)),
        sa.Column('project_id', sa.String(36)),
        sa.Column('compute_id', sa.String(36)),
        sa.Column('topology', sa.String(50)),
        sa.Column('role', sa.String(50)),
        sa.Column('last_udp_update', sa.DateTime(), nullable=True),
        sa.Column('status', sa.String(36), default='ACTIVE', nullable=False),
        sa.Column(u'created_at', sa.DateTime(), nullable=True),
        sa.Column(u'updated_at', sa.DateTime(), nullable=True),
        sa.Column('partition_name', sa.String(14), nullable=True)
    )


def downgrade():
    op.drop_table('vthunders')



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from datetime import datetime
import signal
import sys
import threading
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_reports import guru_meditation_report as gmr

from a10_octavia.cmd import service
from a10_octavia.controller.housekeeping import house_keeping
from octavia import version

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

spare_amp_thread_event = threading.Event()
db_cleanup_thread_event = threading.Event()
write_memory_thread_event = threading.Event()


def spare_amphora_check():
    """Initiates spare amp check with respect to configured interval."""

    # Read the interval from CONF
    interval = CONF.a10_house_keeping.spare_check_interval
    LOG.info("Spare check interval is set to %d sec", interval)
    spare_amp = house_keeping.SpareAmphora()
    while not spare_amp_thread_event.is_set():
        LOG.debug("Initiating spare amphora check...")
        try:
            spare_amp.spare_check()
        except Exception as e:
            LOG.debug('spare_amphora caught the following exception and '
                      'is restarting: {}'.format(e))
        spare_amp_thread_event.wait(interval)


def db_cleanup():
    """Perform db cleanup for old resources."""
    # Read the interval from CONF
    interval = CONF.a10_house_keeping.cleanup_interval
    LOG.info("DB cleanup interval is set to %d sec", interval)
    LOG.info('Amphora expiry age is %s seconds',
             CONF.a10_house_keeping.amphora_expiry_age)
    LOG.info('Load balancer expiry age is %s seconds',
             CONF.a10_house_keeping.load_balancer_expiry_age)

    db_cleanup = house_keeping.DatabaseCleanup()
    while not db_cleanup_thread_event.is_set():
        LOG.info("Initiating the cleanup of old resources...")
        try:
            db_cleanup.delete_old_amphorae()
            db_cleanup.cleanup_load_balancers()
        except Exception as e:
            LOG.info('db_cleanup caught the following exception and '
                     'is restarting: {}'.format(e))
        db_cleanup_thread_event.wait(interval)


def write_memory():
    """Performs write memory for all thunders"""
    if CONF.a10_house_keeping.use_periodic_write_memory == 'enable':
        interval = CONF.a10_house_keeping.write_mem_interval
        write_memory_perform = house_keeping.WriteMemory()
        LOG.info("Write Memory interval set to %s seconds", interval)
        while not write_memory_thread_event.is_set():
            LOG.info("Initiating the write memory operation for all thunders...")
            timestamp = datetime.utcnow()
            LOG.debug("Starting write memory thread ar %s", str(timestamp))
            try:
                write_memory_perform.perform_memory_writes()
            except Exception as e:
                LOG.exception('write memory caught the following exception and '
                              ' is restarting: {}'.format(e))
            write_memory_thread_event.wait(interval)
    else:
        LOG.warning("Write memory flag is disabled...")


def _mutate_config(*args, **kwargs):
    LOG.info("Housekeeping recieved HUP signal, mutating config.")
    CONF.mutate_config_files()


def main():
    service.prepare_service(sys.argv)

    gmr.TextGuruMeditation.setup_autorun(version)

    timestamp = datetime.utcnow()
    LOG.info("Starting house keeping at %s", str(timestamp))

    # Thread to perform spare amphora check
    spare_amp_thread = threading.Thread(target=spare_amphora_check)
    spare_amp_thread.daemon = True
    spare_amp_thread.start()

    # Thread to perform db cleanup
    db_cleanup_thread = threading.Thread(target=db_cleanup)
    db_cleanup_thread.daemon = True
    db_cleanup_thread.start()

    # Thread to perform write memory
    write_memory_thread = threading.Thread(target=write_memory)
    write_memory_thread.daemon = True
    write_memory_thread.start()

    signal.signal(signal.SIGHUP, _mutate_config)

    # Try-Exception block should be at the end to gracefully exit threads
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        LOG.info("Attempting to gracefully terminate House-Keeping")
        spare_amp_thread_event.set()
        db_cleanup_thread_event.set()
        write_memory_thread_event.set()
        spare_amp_thread.join()
        db_cleanup_thread.join()
        write_memory_thread.join()
        LOG.info("House-Keeping process terminated")



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client.errors as acos_errors
from oslo_config import cfg
from oslo_log import log as logging
from requests import exceptions
from taskflow import task
import time

from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator_for_revert
from a10_octavia.controller.worker.tasks import utils


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class LoadBalancerParent(object):

    def set(self, set_method, loadbalancer, flavor_data=None, **kwargs):
        config_args = {
            'arp_disable': CONF.slb.arp_disable,
            'port_list': kwargs.get('port_list'),
            'vrid': CONF.slb.default_virtual_server_vrid,
            'virtual_server': utils.meta(loadbalancer, 'virtual_server', {})
        }

        status = self.axapi_client.slb.UP
        if not loadbalancer.provisioning_status:
            status = self.axapi_client.slb.DOWN
        config_args['status'] = status

        desc = loadbalancer.description
        if not desc:
            desc = None
        elif str(desc).isspace() or not str(desc):
            desc = ""
        else:
            if len(desc) > 61:
                LOG.warning('Description length exceeds the allowed character limit.'
                            ' Trimming the description length')
            desc = '"{}"'.format(desc[0:61])
        config_args['description'] = desc

        if flavor_data:
            virtual_server_flavor = flavor_data.get('virtual_server')
            if virtual_server_flavor:
                name_exprs = virtual_server_flavor.get('name_expressions')
                if name_exprs:
                    del virtual_server_flavor['name_expressions']
                virtual_server_flavor.update(utils.parse_name_expressions(
                    loadbalancer.name, name_exprs))
                config_args['virtual_server'].update(virtual_server_flavor)

        set_method(loadbalancer.id, loadbalancer.vip.ip_address, **config_args)


class CreateVirtualServerTask(LoadBalancerParent, task.Task):
    """Task to create a virtual server"""

    @axapi_client_decorator
    def execute(self, loadbalancer, vthunder, flavor_data=None):
        try:
            self.set(self.axapi_client.slb.virtual_server.create, loadbalancer, flavor_data)
            LOG.debug("Successfully created load balancer: %s", loadbalancer.id)
        except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
            LOG.exception("Failed to created load balancer: %s", loadbalancer.id)
            raise e

    @axapi_client_decorator_for_revert
    def revert(self, loadbalancer, vthunder, flavor_data=None, *args, **kwargs):
        try:
            LOG.warning("Reverting creation of load balancer: %s", loadbalancer.id)
            self.axapi_client.slb.virtual_server.delete(loadbalancer.id)
        except exceptions.ConnectionError:
            LOG.exception(
                "Failed to connect A10 Thunder device: %s", vthunder.ip)
        except Exception as e:
            LOG.exception("Failed to revert creation of load balancer: %s due to %s",
                          loadbalancer.id, str(e))


class DeleteVirtualServerTask(task.Task):
    """Task to delete a virtual server"""

    @axapi_client_decorator
    def execute(self, loadbalancer, vthunder):
        if vthunder:
            try:
                self.axapi_client.slb.virtual_server.delete(loadbalancer.id)
                LOG.debug("Successfully deleted load balancer: %s", loadbalancer.id)
            except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
                LOG.exception("Failed to delete load balancer: %s", loadbalancer.id)
                raise e


class UpdateVirtualServerTask(LoadBalancerParent, task.Task):
    """Task to update a virtual server"""

    @axapi_client_decorator
    def execute(self, loadbalancer, vthunder, flavor_data=None):
        try:
            port_list = self.axapi_client.slb.virtual_server.get(
                loadbalancer.id)['virtual-server'].get('port-list')
            self.set(self.axapi_client.slb.virtual_server.replace, loadbalancer,
                     port_list=port_list, flavor_data=flavor_data)
            LOG.debug("Successfully updated load balancer: %s", loadbalancer.id)
        except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
            LOG.exception("Failed to update load balancer: %s", loadbalancer.id)
            raise e


class WaitVirtualServerReadyOnBlade(LoadBalancerParent, task.Task):
    """Task to wait vBlade get virtual-server configuration from vMaster"""

    @axapi_client_decorator
    def execute(self, loadbalancer, vthunder):
        attempts = CONF.a10_controller_worker.amp_vcs_retries
        while attempts >= 0:
            try:
                attempts = attempts - 1
                self.axapi_client.slb.virtual_server.get(loadbalancer.id)
                break
            except exceptions.ReadTimeout:
                # Don't retry for ReadTimout, since acos-client already already have
                # tries and timeout for axapi request. And it will take very long to
                # response this error.
                if attempts < 0:
                    # Don't raise, LB creation is success just not sync. to vBlade yet
                    break
            except Exception:
                if attempts < 0:
                    # Don't raise, LB creation is success just not sync. to vBlade yet
                    break
                time.sleep(CONF.a10_controller_worker.amp_vcs_wait_sec)



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from a10_octavia.db import api as db_apis
from a10_octavia.db import repositories as repo


class VThunderDB(object):

    def __init__(self, **kwargs):
        self.vthunder_repo = repo.VThunderRepository()

    def create_vthunder(
            self,
            project_id,
            device_name,
            username,
            password,
            ip_address,
            undercloud=None,
            axapi_version=30):

        if axapi_version == 2.1:
            axapi_version = 21
        else:
            axapi_version = 30

        amphora_id = uuidutils.generate_uuid()
        vthunder_id = uuidutils.generate_uuid()

        if undercloud == 'True' or undercloud == 'true':
            undercloud = True
        else:
            undercloud = False

        db_session = db_apis.get_session()
        self.vthunder_repo.create(
            db_session,
            vthunder_id=vthunder_id,
            amphora_id=amphora_id,
            project_id=project_id,
            device_name=device_name,
            username=username,
            password=password,
            ip_address=ip_address,
            undercloud=undercloud,
            axapi_version=axapi_version)
        db_apis.close_session(db_session)

        print("vThunder entry created successfully.")

    def update_vthunder(
            self,
            id,
            project_id,
            device_name,
            username,
            password,
            ip_address,
            undercloud=None,
            axapi_version=30):
        if axapi_version == 2.1:
            axapi_version = 21
        else:
            axapi_version = 30

        if undercloud == 'True' or undercloud == 'true':
            undercloud = True
        else:
            undercloud = False

        db_session = db_apis.get_session()
        self.vthunder_repo.update(db_session, id, project_id=project_id,
                                  device_name=device_name, username=username,
                                  password=password, ip_address=ip_address,
                                  undercloud=undercloud, axapi_version=axapi_version)
        db_apis.close_session(db_session)

        print("vThunder entry updated successfully.")

    def delete_vthunder(self, vthunderid):
        db_session = db_apis.get_session()
        self.vthunder_repo.delete(db_session, id=vthunderid)
        db_apis.close_session(db_session)
        print("vThunder entry deleted successfully.")



"""Added NAT Pool table

Revision ID: a41260fb4ccd
Revises: 05b1446c7f20
Create Date: 2020-12-21 16:52:42.793854

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = 'a41260fb4ccd'
down_revision = '05b1446c7f20'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('nat_pool',
    sa.Column('id', sa.String(length=64), nullable=False),
    sa.Column('name', sa.String(length=64), nullable=False),
    sa.Column('subnet_id', sa.String(length=64), nullable=False),
    sa.Column('start_address', sa.String(length=64), nullable=False),
    sa.Column('end_address', sa.String(length=64), nullable=False),
    sa.Column('member_ref_count', sa.Integer(), nullable=True),
    sa.Column('port_id', sa.String(length=64), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name', 'subnet_id', name='unique_name_subnet_id')
    )


def downgrade():
    op.drop_table('nat_pool')



"""vrid_table_extend_owner_size

Revision ID: 1722c3166462
Revises: 873ee83aef63
Create Date: 2021-06-25 18:36:26.122495

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1722c3166462'
down_revision = '873ee83aef63'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('vrid', 'owner',
            existing_type=sa.String(128), nullable=False)


def downgrade():
    op.alter_column('vrid', 'owner',
            existing_type=sa.String(36), nullable=False)



"""Add subnet_id column to vrid table

Revision ID: 05b1446c7f20
Revises: ea8bfcbd654c
Create Date: 2020-09-10 07:08:19.928899

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '05b1446c7f20'
down_revision = '4028e5f7a198'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('vrid', sa.Column('subnet_id', sa.String(36), nullable=False))


def downgrade():
    op.drop_column('vrid', 'subnet_id')



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

from oslo_log import log as logging

from a10_octavia.common.defaults import DEFAULT
from a10_octavia.common import exceptions
from a10_octavia.etc import config as blank_config
from a10_octavia.etc import defaults

# This is ConfigParser pre-Python3
if sys.version_info < (3,):
    import ConfigParser as ini
else:
    import configparser as ini


LOG = logging.getLogger(__name__)


class ConfigModule(object):
    def __init__(self, d, provider=None):
        self.__dict__.update(d)

        if provider is None or 'providers' not in d or provider not in d['providers']:
            return None

        for k, v in d['providers'][provider].items():
            if isinstance(v, dict):
                if k not in self.__dict__:
                    self.__dict__[k] = {}
                self.__dict__[k].update(v)
            else:
                self.__dict__[k] = v

    @classmethod
    def load(cls, path, provider=None):
        d = dict()
        return ConfigModule(d, provider=provider)


class A10Config(object):

    def __init__(self, config_dir=None, config=None, provider=None):
        if config is not None:
            self._config = config
            self._load_config()
            return

        self._conf = None
        self._config_dir = self._find_config_dir(config_dir)
        self._config_path = os.path.join(self._config_dir, "a10-octavia.conf")

        try:
            self._config = ConfigModule.load(self._config_path, provider=provider)
            config_parser = ini.ConfigParser(defaults=DEFAULT)
            config_parser.read(self._config_path)
            self._conf = config_parser
        except IOError:
            LOG.error("A10Config could not find %s", self._config_path)
            self._conf = ini.ConfigParser(defaults=DEFAULT)
            self._config = blank_config

        self._config.octavia_conf_dir = '/etc/octavia/'
        self._load_config()

    def get_conf(self):
        return self._conf

    def _find_config_dir(self, config_dir):
        # Look for config in the virtual environment
        # virtualenv puts the original prefix in sys.real_prefix
        # pyenv puts it in sys.base_prefix
        venv_d = os.path.join(sys.prefix, 'etc/a10')
        has_prefix = (hasattr(sys, 'real_prefix') or hasattr(sys, 'base_prefix'))

        env_override = os.environ.get('A10_CONFIG_DIR', None)
        if config_dir is not None:
            directory = config_dir
        elif env_override is not None:
            directory = env_override
        elif has_prefix and os.path.exists(venv_d):
            directory = venv_d
        else:
            directory = '/etc/a10'

        return directory

    def _load_config(self):
        # Global defaults
        for k, v in defaults.GLOBAL_DEFAULTS.items():
            if not hasattr(self._config, k):
                LOG.debug("setting global default %s=%s", k, v)
                setattr(self._config, k, v)
            else:
                LOG.debug("global setting %s=%s", k, getattr(self._config, k))

        # Setup db foo
        if self._config.database_connection is None:
            self._config.database_connection = self._get_octavia_db_string()

        if self._config.keystone_auth_url is None:
            self._config.keystone_auth_url = self.get_octavia_conf('keystone_authtoken', 'auth_uri')

    def get_octavia_conf(self, section, option):
        octavia_conf_dir = os.environ.get('OCTAVIA_CONF_DIR', self._config.octavia_conf_dir)
        octavia_conf = '%s/octavia.conf' % octavia_conf_dir

        if os.path.exists(octavia_conf):
            LOG.debug("found octavia.conf file in /etc")
            n = ini.ConfigParser()
            n.read(octavia_conf)
            try:
                return n.get(section, option)
            except (ini.NoSectionError, ini.NoOptionError):
                pass
        else:
            raise Exception('FatalError: Octavia config directoty could not be found.')
            LOG.error("A10Config could not find %s", self._config_path)

    def _get_octavia_db_string(self):
        db_connection_url = self.get_octavia_conf('database', 'connection')

        if db_connection_url is None:
            raise exceptions.NoDatabaseURL()

        LOG.debug("Using %s as db connect string", db_connection_url)
        return db_connection_url

    def get(self, key):
        return getattr(self._config, key)
